import enum
import numpy as np


class IntValueEnum(int, enum.Enum):
    """
    An enumerator base class that allows the setting of both a value and label.
    IntEnum(value) returns the as expected

    Based on https://stackoverflow.com/a/68400507
    """

    def __new__(cls, value, number: int):
        obj = int.__new__(cls, number)
        obj._value_ = value
        obj.number = number
        return obj

    def __str__(self):
        return f"{self.name} ({self.value.__name__})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.value.__name__}, {self.number})"

    def __format__(self, format_spec):
        return str(self)

    @classmethod
    def from_number(cls, number: int, strict=False):
        for obj in cls:
            if obj.number == number:
                return obj
        if strict:
            raise ValueError(f"{cls.__name__} has no value matching {number}")
        return None


class XrmDataTypes(IntValueEnum):
    """Integer Enumerator for Xradia's XrmDataTypes"""
    # XRM_BIT = None, 1
    XRM_CHAR = np.int8, 2
    XRM_UNSIGNED_CHAR = np.uint8, 3
    XRM_SHORT = np.int16, 4
    XRM_UNSIGNED_SHORT = np.uint16, 5
    XRM_INT = np.int32, 6
    XRM_UNSIGNED_INT = np.uint32, 7
    XRM_LONG = np.int64, 8
    XRM_UNSIGNED_LONG = np.uint64, 9
    XRM_FLOAT = np.float32, 10
    XRM_DOUBLE = np.float64, 11
    XRM_STRING = np.str_, 12
    # XRM_DATATYPE_SIZE = None, 13


class AnnotationTypes(enum.Enum):
    ANN_LINE = 0
    ANN_RECT = 1
    ANN_ELLIPSE = 2
    ANN_CIRCLE = 3
    ANN_TEXT = 4  # Doesn't work in XRMDataExplorer
    ANN_POLYGON = 5
    ANN_GROUP = 6
    ANN_LINE_PLOT = 7
    ANN_RULE = 8
    ANN_ANGLE = 9
    ANN_MOUSEPOINT = 10
    ANN_CONE_ANGLE_LINE = 11
    ANN_HIGHLIGHTER = 12
    ANN_POLYLINE = 13
    ANN_FREE_HAND_SKETCH = 14
    ANN_SIZE = 15



import logging

from pathlib import Path
from typing import Optional
import numpy as np
from numpy.typing import DTypeLike

from ..utils.metadata import create_ome_metadata, dtype_dict
from ..utils.file_handler import manual_save


class SaveMixin:
    def save_images(
        self,
        filepath: Optional[Path] = None,
        datatype: Optional[DTypeLike] = None,
        shifts: bool = False,
        flip: bool = False,
        clear_images: bool = False,
        mkdir: bool = False,
        save_annotations: bool = True,
        annotated_path: Optional[Path] = None,
        strict: Optional[bool] = None,
    ) -> bool:
        """Saves images (if available) returning True if successful."""
        if strict is None:
            strict = self.strict
        try:
            if filepath is None:
                if self.path is None:
                    raise ValueError(
                        "An output filepath must be given if an input path was not given."
                    )
                filepath = self.path.resolve().with_suffix(".ome.tiff")
            if not self.referenced:
                logging.info("Saving without reference")

            im = self.get_output(
                load=True, shifts=shifts, flip=flip, clear_images=clear_images
            )
            if im is None:
                raise AttributeError("Cannot save image as no image has been loaded.")
            if datatype is not None:
                datatype = np.dtype(datatype)
                if datatype.name not in dtype_dict:
                    datatype = None
                    logging.warning(
                        "Invalid data type '%s', must be %s or None. Defaulting to saving as %s",
                        datatype,
                        ", ".join(dtype_dict.keys()),
                        str(im.dtype),
                    )
            metadata = self.create_metadata(filepath)

            if mkdir:
                filepath.parent.mkdir(parents=True, exist_ok=True)

            manual_save(filepath, im, datatype, metadata)
            if (
                save_annotations
                and hasattr(self, "annotate")
                and self.annotated_image is not None
            ):
                if annotated_path is None:
                    # Generate default path
                    filename = filepath.name
                    if filename.lower().endswith(".ome.tiff"):
                        # Special case for ome.tiff
                        stem, suffix = filename.rsplit(".ome.", 1)
                    else:
                        stem, suffix = filename.rsplit(".", 1)
                    annotated_path = filepath.parent / f"{stem}_Annotated.{suffix}"
                self.save_annotations(annotated_path, mkdir=mkdir, strict=strict)
            return True
        except Exception:
            logging.error("Saving failed", exc_info=not strict)
            if strict:
                raise
            return False

    def create_metadata(self, filepath: Path):
        return create_ome_metadata(self, filepath.stem)



import logging
import typing


def txrm_property(
    function: typing.Optional[typing.Callable] = None,
    fallback: typing.Any = None,
    **kwargs,
):
    class TxrmProperty:
        def __init__(self, fn, fallback):
            self.fn = fn
            self.fallback = fallback
            self.hidden_var_str = f"_{self.fn.__name__}"

        def __get__(self, obj, object_type=None):
            try:
                if getattr(obj, self.hidden_var_str, None) is None:
                    try:
                        setattr(obj, self.hidden_var_str, self.fn(obj))
                    except Exception:
                        if not obj.file_is_open:
                            raise IOError(
                                f"Cannot get {self.fn.__name__} while file is closed"
                            )
                        raise
            except Exception:
                if obj.strict:
                    raise
                logging.error(
                    "Failed to get property '%s'",
                    self.fn.__name__,
                    exc_info=True,
                )
            return getattr(obj, self.hidden_var_str, self.fallback)

        def __set__(self, obj, value):
            raise AttributeError(f"{self.fn.__name__} cannot be set.")

        def __delete__(self, obj):
            setattr(obj, self.hidden_var_str, None)

    if function:
        return TxrmProperty(function)
    else:

        def wrapper(function):
            return TxrmProperty(function, fallback, **kwargs)

        return wrapper



import unittest
from unittest.mock import MagicMock, patch

import os
import pathlib
from pathlib import Path

from txrm2tiff.utils.shortcut_creation import create_Windows_shortcut, _create_lnk_file


class TestSetupFunctions(unittest.TestCase):
    @patch.object(pathlib.Path, "exists", MagicMock(return_value=False))
    @patch("logging.error")
    def test_setup_windows_shortcut_function_called(self, mocked_error):
        with patch(
            "txrm2tiff.utils.shortcut_creation._create_lnk_file", MagicMock()
        ) as mocked_shortcut_creation:
            create_Windows_shortcut()
            if os.name == "nt":
                mocked_shortcut_creation.assert_called_once_with(
                    Path(os.environ["HOMEDRIVE"])
                    / os.environ["HOMEPATH"]
                    / "Desktop"
                    / "txrm2tiff.lnk"
                )
            else:
                mocked_error.assert_called_once_with(
                    "This command is only valid on Windows installations."
                )

    def test_setup_windows_shortcut_test_created(self):
        # Only run this test if on Windows
        if os.name == "nt":
            test_shortcut_path = Path(".") / "test_shortcut.lnk"
            _create_lnk_file(test_shortcut_path)
            link_created = test_shortcut_path.exists()
            if link_created:
                os.remove(test_shortcut_path)
            self.assertTrue(link_created)



import logging
import typing
import numpy as np
import olefile as of

from ..xradia_properties import XrmDataTypes as XDT

from . import general


def extract_image_dtype(
    ole: of.OleFileIO, key_part: str, strict: bool = False, **kwargs
) -> typing.Optional[XDT]:
    key = f"{key_part}/DataType"
    integer_list = general.read_stream(ole, key, XDT.XRM_INT, strict)
    if not integer_list:
        return None
    return XDT.from_number(integer_list[0], strict)


def extract_single_image(
    ole: of.OleFileIO,
    image_num: int,
    numrows: int,
    numcols: int,
    strict: bool = False,
    **kwargs,
) -> np.ndarray:
    # Read the images - They are stored in the txrm as ImageData1 ...
    # Each folder contains 100 images 1-100, 101-200
    img_key = f"ImageData{int(np.ceil(image_num / 100.0))}/Image{image_num}"
    if not ole.exists(img_key):
        if strict:
            raise KeyError("Stream '%s' does not exist" % img_key)
        return np.asarray([])
    else:
        img_stream_bytes = ole.openstream(img_key).getvalue()
        image_dtype = extract_image_dtype(ole, "ImageInfo", strict=False)
        if image_dtype is not None:
            try:
                imgdata = general.get_stream_from_bytes(
                    img_stream_bytes, dtype=image_dtype.value
                )
            except Exception:
                logging.error(
                    "Image could not be extracted using expected dtype '%s'",
                    image_dtype,
                )
                if strict:
                    raise
        else:
            try:
                img_size = numrows * numcols
                imgdata = fallback_image_interpreter(img_stream_bytes, img_size, strict)
            except Exception:
                logging.error("Exception occurred getting %s", img_key, exc_info=True)
                if strict:
                    raise
        imgdata.shape = (numrows, numcols)
        return imgdata


def fallback_image_interpreter(
    stream_bytes: bytes,
    image_size: int,
    strict: bool = False,
):
    stream_length = len(stream_bytes)
    if stream_length == image_size * 2:
        dtype = XDT.XRM_UNSIGNED_SHORT
    elif stream_length == image_size * 4:
        dtype = XDT.XRM_FLOAT
    else:
        logging.error(
            "Unexpected data type with %g bytes per pixel",
            stream_length / image_size,
        )
        msg = "Image is stored as unexpected type. Expecting uint16 or float32."
        if strict:
            raise TypeError(msg)
        logging.error(msg)
        return np.asarray([])
    return np.frombuffer(stream_bytes, dtype=dtype.value)



#!/usr/bin/python
import argparse
import sys
import os
if os.name == "nt":
    sys.stdout.reconfigure(encoding="utf-8")

from .info import __version__

parser = argparse.ArgumentParser(
    prog="txrm2tiff",
    description="Converter of txrm/xrm files to OME tif/tiff files",
    add_help=False,
)
convert_group = parser.add_argument_group("Converter arguments")
convert_group.add_argument(
    "-i",
    "--input",
    dest="input_path",
    default=None,
    type=str,
    action="store",
    help="txrm or xrm file to convert, or directory containing txrm and/or xrm files",
)
convert_group.add_argument(
    "-r",
    "--reference",
    dest="custom_reference",
    default=None,
    type=str,
    action="store",
    help="specify a custom file (tiff, txrm or xrm) for reference (ignores internal reference)",
)
convert_group.add_argument(
    "-o",
    "--output-path",
    dest="output_path",
    default=None,
    action="store",
    help="specify output (can be directory or file) - specified directories that do not exist will be created",
)
convert_group.add_argument(
    "-a",
    "--annotate",
    dest="annotate",
    action="store_true",
    help="save a second annotated copy, if annotations are available",
)
convert_group.add_argument(
    "-f",
    "--flip",
    dest="flip",
    action="store_true",
    help="if specified, the output image will be flipped in the Y-axis with respect to how it displays in XRM Data Explorer",
)
convert_group.add_argument(
    "-d",
    "--datatype",
    dest="data_type",
    default=None,
    action="store",
    choices=["uint16", "float32", "float64"],
    help="specify output data type (default: decides data type from input)",
)
convert_group.add_argument(
    "--apply-shifts",
    dest="apply_shifts",
    default=False,
    action="store_true",
    help="if specified, apply saved shifts (if available)",
)
convert_group.add_argument(
    "--ignore-ref",
    dest="ignore_reference",
    default=False,
    action="store_true",
    help="if specified, ignore any internal reference",
)
convert_group.add_argument(
    "--set-logging",
    dest="logging_level",
    default="info",
    type=str,
    action="store",
    help="pick logging level (options: debug, info, warning, error, critical)",
)

package_group = parser.add_argument_group("Package info")
package_group.add_argument(
    "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
)
package_group.add_argument(
    "-h", "--help", action="help", help="show this help message and exit"
)

subparsers = parser.add_subparsers(
    title="Additional options", help="Additional functionality options"
)
inspect_parser = subparsers.add_parser(name="inspect", add_help=False)
inspect_info_group = inspect_parser.add_argument_group("Inspection options")
inspect_info_group.add_argument(
    "-i",
    "--input",
    dest="input_path",
    default="",
    type=str,
    action="store",
    help="txrm or xrm file to inspect",
)
inspect_info_group.add_argument(
    "-e", "--extra", action="store_true", help="Shows additional file info"
)
inspect_info_group.add_argument(
    "-l", "--list-streams", action="store_true", help="List all streams"
)
inspect_info_group.add_argument(
    "-s", "--inspect-streams", dest="streams", nargs="+", help="Inspect stream"
)
inspect_info_group.add_argument(
    "-h", "--help", action="help", help="show this help message and exit"
)

setup_parser = subparsers.add_parser(name="setup", add_help=False)
setup_info_group = setup_parser.add_argument_group("Setup options")
setup_info_group.add_argument(
    "-w",
    "--windows-shortcut",
    action="store_true",
    help="creates a Windows shortcut on the user's desktop that accepts drag and dropped files (allows batch processing)",
)
setup_info_group.add_argument(
    "-h", "--help", action="help", help="show this help message and exit"
)


def main():
    if "inspect" in sys.argv:
        inspect_args = parser.parse_args(namespace=inspect_parser)
        if inspect_args.input_path:
            from .inspector import Inspector
            from .txrm import open_txrm

            try:
                with open_txrm(
                    inspect_args.input_path,
                    load_images=False,
                    load_reference=False,
                    strict=False,
                ) as txrm:
                    inspector = Inspector(txrm)
                    inspector.inspect(extra=bool(inspect_args.extra))
                    if inspect_args.list_streams:
                        inspector.list_streams()
                    if inspect_args.streams:
                        inspector.inspect_streams(*inspect_args.streams)
                    print(inspector.get_text())
                return
            except Exception as e:
                import traceback

                print(
                    "Exception occurred while inspecting %s:\n\n%s"
                    % (
                        inspect_args.input_path,
                        "".join(
                            traceback.format_exception(type(e), e, e.__traceback__)
                        ),
                    )
                )
            return
        else:
            inspect_parser.print_help()
            return

    if "setup" in sys.argv:
        setup_args = parser.parse_args(namespace=setup_parser)
        if setup_args.windows_shortcut:
            from .utils.shortcut_creation import create_Windows_shortcut

            create_Windows_shortcut()
            return
        else:
            setup_parser.print_help()
            return

    args = parser.parse_args()
    if args.input_path:
        from . import convert_and_save

        convert_and_save(**vars(args))

    else:
        parser.print_help()
        print("\n\n\ntxrm2tiff inspect:\n")
        inspect_parser.print_help()
        print("\n\n\ntxrm2tiff setup:\n")
        setup_parser.print_help()


if __name__ == "__main__":
    main()



from pathlib import Path
import unittest
from unittest.mock import patch, MagicMock, call

from txrm2tiff.txrm.save_mixin import SaveMixin
from txrm2tiff.txrm.annot_mixin import AnnotatorMixin


class SaveAnnotatorMixin(SaveMixin, AnnotatorMixin):
    ...


class TestSaver(unittest.TestCase):
    @patch("pathlib.Path.mkdir", MagicMock())
    @patch("txrm2tiff.txrm.save_mixin.create_ome_metadata")
    def test_create_metadata(self, mocked_metadata_creator):
        filepath = Path("path/to/file.ext")
        saver = SaveMixin()
        saver.create_metadata(filepath)
        mocked_metadata_creator.assert_called_once_with(saver, filepath.stem)

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.annot_mixin.manual_annotation_save")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images(
        self, mocked_mkdir, mocked_manual_save, mocked_ann_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ext")
        dtype = None
        shifts = True
        flip = True
        clear_images = True
        mkdir = False

        image = "image"
        metadata = "metadata"

        saver = SaveMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.referenced = False
        saver.annotated_image = None

        self.assertTrue(
            saver.save_images(
                filepath, dtype, shifts, flip, clear_images, mkdir, strict=True
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_not_called()
        mocked_manual_save.assert_called_once_with(filepath, image, dtype, metadata)
        mocked_ann_save.assert_not_called()

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_mkdir(
        self, mocked_mkdir, mocked_manual_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ext")
        dtype = None
        shifts = True
        flip = True
        clear_images = True
        mkdir = True

        image = "image"
        metadata = "metadata"

        saver = SaveMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.save_annotations = MagicMock()
        saver.referenced = False
        saver.annotated_image = None

        self.assertTrue(
            saver.save_images(
                filepath, dtype, shifts, flip, clear_images, mkdir, strict=True
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_called_once()
        mocked_manual_save.assert_called_once_with(filepath, image, dtype, metadata)
        saver.save_annotations.assert_not_called()

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_creates_ome_tiff(
        self, mocked_mkdir, mocked_manual_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ext")
        dtype = None
        shifts = True
        flip = True
        clear_images = True
        mkdir = False

        image = "image"
        metadata = "metadata"

        saver = SaveMixin()

        mocked_create_metadata.return_value = metadata
        saver.path = filepath
        saver.get_output = MagicMock(return_value=image)
        saver.save_annotations = MagicMock()
        saver.referenced = False
        saver.annotated_image = None

        self.assertTrue(
            saver.save_images(
                None, dtype, shifts, flip, clear_images, mkdir, strict=True
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_not_called()
        mocked_manual_save.assert_called_once_with(
            filepath.resolve().with_suffix(".ome.tiff"), image, dtype, metadata
        )
        saver.save_annotations.assert_not_called()

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.annot_mixin.manual_annotation_save")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_and_annotations_simple_ext(
        self, mocked_mkdir, mocked_manual_save, mocked_ann_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ext")
        dtype = None
        shifts = False
        flip = True
        clear_images = True
        mkdir = True

        image = "image"
        metadata = "metadata"
        annotated_image = "annotated"

        saver = SaveAnnotatorMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.referenced = False
        saver.annotated_image = annotated_image

        self.assertTrue(
            saver.save_images(
                filepath, dtype, shifts, flip, clear_images, mkdir, strict=True
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_has_calls([call(parents=True, exist_ok=True)] * 2)
        mocked_manual_save.assert_called_once_with(filepath, image, dtype, metadata)
        mocked_ann_save.assert_called_once_with(
            Path("path/to/file_Annotated.ext"), annotated_image
        )

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.annot_mixin.manual_annotation_save")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_and_annotations_tiff(
        self, mocked_mkdir, mocked_manual_save, mocked_ann_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ome.tiff")
        dtype = None
        shifts = False
        flip = True
        clear_images = True
        mkdir = False

        image = "image"
        metadata = "metadata"
        annotated_image = "annotated"

        saver = SaveAnnotatorMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.referenced = False
        saver.annotated_image = annotated_image

        self.assertTrue(
            saver.save_images(
                filepath, dtype, shifts, flip, clear_images, mkdir, strict=True
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_not_called()
        mocked_manual_save.assert_called_once_with(filepath, image, dtype, metadata)
        mocked_ann_save.assert_called_once_with(
            Path("path/to/file_Annotated.tiff"), annotated_image
        )

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.annot_mixin.manual_annotation_save")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_and_annotations_specific_filepath(
        self, mocked_mkdir, mocked_manual_save, mocked_ann_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ome.tiff")
        dtype = None
        shifts = False
        flip = True
        clear_images = True
        mkdir = True

        image = "image"
        metadata = "metadata"
        annotated_image = "annotated"

        requested_filepath = Path("path/to/file_Annotated.tiff")

        saver = SaveAnnotatorMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.referenced = False
        saver.annotated_image = annotated_image

        self.assertTrue(
            saver.save_images(
                filepath,
                dtype,
                shifts,
                flip,
                clear_images,
                mkdir,
                strict=True,
                annotated_path=requested_filepath,
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )
        mocked_mkdir.assert_has_calls([call(parents=True, exist_ok=True)] * 2)
        mocked_manual_save.assert_called_once_with(filepath, image, dtype, metadata)
        mocked_ann_save.assert_called_once_with(requested_filepath, annotated_image)

    @patch.object(SaveMixin, "create_metadata")
    @patch("txrm2tiff.txrm.save_mixin.manual_save")
    @patch("pathlib.Path.mkdir")
    def test_save_images_returns_False_without_image(
        self, mocked_mkdir, mocked_manual_save, mocked_create_metadata
    ):
        filepath = Path("path/to/file.ext")
        dtype = None
        shifts = False
        flip = True
        clear_images = True
        mkdir = True

        image = None
        metadata = "metadata"
        annotated_image = "annotated"

        saver = SaveMixin()

        mocked_create_metadata.return_value = metadata
        saver.get_output = MagicMock(return_value=image)
        saver.save_annotations = MagicMock()
        saver.referenced = False
        saver.annotated_image = annotated_image

        self.assertFalse(
            saver.save_images(
                filepath, dtype, shifts, flip, clear_images, mkdir, strict=False
            )
        )

        saver.get_output.assert_called_once_with(
            load=True, shifts=shifts, flip=flip, clear_images=clear_images
        )

        mocked_mkdir.assert_not_called()
        mocked_manual_save.assert_not_called()
        saver.save_annotations.assert_not_called()



import logging
import numpy as np
from oxdls import OMEXML, DO_XYZCT, PT_UINT16, PT_UINT16, PT_FLOAT, PT_DOUBLE

from ..txrm import abstract

dtype_dict = {"uint16": PT_UINT16, "float32": PT_FLOAT, "float64": PT_DOUBLE}




def create_ome_metadata(txrm: abstract.AbstractTxrm, filename: str = None) -> OMEXML:
    # Get image shape
    # number of frames (T in tilt series), Y, X:
    shape = txrm.output_shape

    # Get metadata variables from ole file:
    exposures = txrm.exposures.copy()

    pixel_size = txrm.image_info.get("PixelSize", (0,))[0] * 1.0e3  # micron to nm

    x_positions = [
        coord * 1.0e3 for coord in txrm.image_info["XPosition"]
    ]  # micron to nm
    y_positions = [
        coord * 1.0e3 for coord in txrm.image_info["YPosition"]
    ]  # micron to nm
    z_positions = [
        coord * 1.0e3 for coord in txrm.image_info["ZPosition"]
    ]  # micron to nm

    ox = OMEXML()

    image = ox.image()
    image.set_ID("Image:0")
    image.set_AcquisitionDate(txrm.datetimes[0].isoformat())  # formatted as: "yyyy-mm-ddThh:mm:ss"
    if filename is not None:
        image.set_Name(filename)

    pixels = image.Pixels
    pixels.DimensionOrder = DO_XYZCT
    pixels.set_ID("Pixels:0")

    pixels.set_SizeX(shape[2])
    pixels.set_SizeY(shape[1])
    pixels.set_SizeC(1)
    pixels.set_SizeZ(shape[0])

    pixels.set_SizeT(1)

    # Physical size of a pixel
    pixels.set_PhysicalSizeX(pixel_size)
    pixels.set_PhysicalSizeXUnit("nm")
    pixels.set_PhysicalSizeY(pixel_size)
    pixels.set_PhysicalSizeYUnit("nm")
    pixels.set_PhysicalSizeZ(1)
    pixels.set_PhysicalSizeZUnit("reference frame")

    pixels.set_tiffdata_count(shape[0])  # tiffdata must be created before planes
    pixels.set_plane_count(shape[0])

    channel = pixels.Channel(0)
    channel.set_ID("Channel:0:0")
    channel.set_Name("C:0")

    if txrm.is_mosaic:
        mosaic_columns, mosaic_rows = txrm.mosaic_dims
        # Calculates:
        # - Mean exposure, throwing away any invalid 0 values
        # - The centre of the stitched mosaic image (as opposed to the centre of a single tile)
        # Both should be returned as a list to reduce changes to the next section.
        valid_idxs = np.nonzero(exposures)[0]
        exposures = [np.mean(np.asarray(exposures)[valid_idxs])]
        # The mosaic centre is found by taking the first x & y positions (centre of the first tile,
        # which is the bottom-left in the mosaic), taking away the distance between this and the
        # bottom-left corner, then adding the distance to the centre of the mosaic (calculated using pixel size).
        #
        # More verbosely:
        # The physical size of the stitched mosaic is divided by the rows/columns (columns for x, rows for y).
        # This finds the physical size of a single tile. This is then halved, finding in the physical
        # distance (x, y) from between a corner and the centre of a tile. Then this distance is taken from the
        # (x, y) stage coordinates of the first tile to get the stage coordinates of the bottom-left of the mosaic.
        # Half the physical size of the stitched mosaic is added to this, resulting in the in stage coordinates
        # of the mosaic centre.
        x_positions = [
            x_positions[0]
            + (1.0 - 1.0 / mosaic_columns) * (pixel_size * shape[2] / 2.0)
        ]
        y_positions = [
            y_positions[0] + (1.0 - 1.0 / mosaic_rows) * (pixel_size * shape[1] / 2.0)
        ]
        z_positions = [np.mean(np.asarray(z_positions)[valid_idxs])]  # Average Z for a stitched mosaic
        # # NOTE: the number of mosaic rows & columns and the pixel size are all written before acquisition but
        # the xy positions are written during, so only the first frame can be relied upon to have an xy
        # position.

    # Run checks to make sure the value lists are long enough
    exp_len_diff = shape[0] - len(exposures)
    if exp_len_diff > 0:
        logging.error(
            "Not enough exposure values for each plane (%i vs %i). Adding zeros to the later planes.",
            len(exposures),
            shape[0],
        )
        for _ in range(exp_len_diff):
            exposures.append(0)

    x_len_diff = shape[0] - len(x_positions)
    if x_len_diff > 0:
        logging.error(
            "Not enough x values for each plane (%i vs %i). Adding zeros to the later planes.",
            len(x_positions),
            shape[0],
        )
        for _ in range(x_len_diff):
            x_positions.append(0)

    y_len_diff = shape[0] - len(y_positions)
    if y_len_diff > 0:
        logging.error(
            "Not enough y values for each plane (%i vs %i). Adding zeros to the later planes.",
            len(y_positions),
            shape[0],
        )
        for _ in range(y_len_diff):
            y_positions.append(0)

    # Add plane/tiffdata for each plane in the stack
    for count in range(shape[0]):
        tiffdata = pixels.tiffdata(count)
        tiffdata.set_FirstC(0)
        tiffdata.set_FirstZ(count)
        tiffdata.set_FirstT(0)
        tiffdata.set_IFD(count)
        tiffdata.set_plane_count(1)

        plane = pixels.Plane(count)
        plane.set_ExposureTime(exposures[count])
        plane.set_TheZ(count)
        plane.set_TheC(0)
        plane.set_TheT(0)

        plane.set_PositionXUnit("nm")
        plane.set_PositionYUnit("nm")
        plane.set_PositionZUnit("reference frame")
        plane.set_PositionX(x_positions[count])
        plane.set_PositionY(y_positions[count])
        plane.set_PositionZ(z_positions[count])

    return ox



from .enums import XrmDataTypes


image_info_dict = {
    "ImageInfo/Angles": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/CameraBinning": XrmDataTypes.XRM_INT,
    "ImageInfo/CameraName": XrmDataTypes.XRM_STRING,
    "ImageInfo/CameraNumberOfFramesPerImage": XrmDataTypes.XRM_INT,
    "ImageInfo/CameraTemperature": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/CamPixelSize": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/Current": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/Date": XrmDataTypes.XRM_STRING,
    "ImageInfo/Energy": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/ExpTime": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/ExpTimes": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/HorizontalBin": XrmDataTypes.XRM_INT,
    "ImageInfo/VerticalalBin": XrmDataTypes.XRM_INT,
    "ImageInfo/ImageHeight": XrmDataTypes.XRM_INT,
    "ImageInfo/ImageWidth": XrmDataTypes.XRM_INT,
    "ImageInfo/ImagesPerProjection": XrmDataTypes.XRM_INT,
    "ImageInfo/ImagesTaken": XrmDataTypes.XRM_UNSIGNED_INT,
    "ImageInfo/IonChamberCurrent": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/MosaicFastAxis": XrmDataTypes.XRM_INT,
    "ImageInfo/MosaicSlowAxis": XrmDataTypes.XRM_INT,
    "ImageInfo/MosiacColumns": XrmDataTypes.XRM_INT,
    "ImageInfo/MosiacRows": XrmDataTypes.XRM_INT,
    "ImageInfo/MosiacMode": XrmDataTypes.XRM_INT,
    "ImageInfo/NoOfImages": XrmDataTypes.XRM_INT,
    "ImageInfo/ObjectiveName": XrmDataTypes.XRM_STRING,
    "ImageInfo/OpticalMagnification": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/PixelSize": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/Temperature": XrmDataTypes.XRM_INT,
    "ImageInfo/XPosition": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/XrayCurrent": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/XrayMagnification": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/YPosition": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/ZPosition": XrmDataTypes.XRM_FLOAT,
    "ImageInfo/ZonePlateName": XrmDataTypes.XRM_STRING,
}

reference_data_dict = {
    "ReferenceData/ExpTime": XrmDataTypes.XRM_FLOAT
    }

ref_image_info_dict = {f"ReferenceData/{k}": v for k, v in image_info_dict.items()}

dtypes_dict = {
    "ImageInfo/DataType": XrmDataTypes.XRM_INT,
    "ReferenceData/ImageInfo/DataType": XrmDataTypes.XRM_INT,
    "ReferenceData/DataType": XrmDataTypes.XRM_INT,
}
annotations_dict = {
    "Annot/TotalAnn": XrmDataTypes.XRM_INT,
}

position_info_dict = {
    "PositionInfo/AxisNames": XrmDataTypes.XRM_STRING,
    "PositionInfo/AxisUnits": XrmDataTypes.XRM_STRING,
    "PositionInfo/MotorPositions": XrmDataTypes.XRM_FLOAT,
    "PositionInfo/TotalAxis": XrmDataTypes.XRM_INT,
}

alignment_dict = {
    "Alignment/EncoderShiftsApplied": XrmDataTypes.XRM_INT,
    "Alignment/MetrologyShiftsApplied": XrmDataTypes.XRM_INT,
    "Alignment/ReferenceShiftsApplied": XrmDataTypes.XRM_INT,
    "Alignment/StageShiftsApplied": XrmDataTypes.XRM_INT,
    "Alignment/StaticRunoutApplied": XrmDataTypes.XRM_INT,
    "Alignment/X-Shifts": XrmDataTypes.XRM_FLOAT, # Pixels
    "Alignment/Y-Shifts": XrmDataTypes.XRM_FLOAT, # Pixels
}

misc = {"exeVersion": XrmDataTypes.XRM_STRING}

streams_dict = {}
streams_dict.update(image_info_dict)
streams_dict.update(reference_data_dict)
streams_dict.update(ref_image_info_dict)
streams_dict.update(dtypes_dict)
streams_dict.update(annotations_dict)
streams_dict.update(position_info_dict)
streams_dict.update(alignment_dict)
streams_dict.update(misc)



from tempfile import TemporaryDirectory
import unittest
from unittest.mock import patch, MagicMock
from parameterized import parameterized
from numpy.testing import assert_array_equal

from pathlib import Path
import numpy as np
from oxdls import OMEXML

from txrm2tiff.utils.file_handler import manual_save
from txrm2tiff.txrm.ref_mixin import ReferenceMixin
from txrm2tiff.txrm.v3 import Txrm3
from txrm2tiff.txrm.v5 import Txrm5


class TestReferencer(unittest.TestCase):
    def test_apply_reference_defaults_with_internal(self):
        referencer = ReferenceMixin()
        referencer.has_reference = True
        with patch.object(
            referencer, "apply_internal_reference"
        ) as mocked_apply_internal:
            referencer.apply_reference()
            mocked_apply_internal.assert_called_once_with(
                overwrite=True, compensate_exposure=True
            )

    def test_apply_reference_defaults_no_internal(self):
        referencer = ReferenceMixin()
        referencer.has_reference = False
        with patch.object(
            referencer, "apply_internal_reference"
        ) as mocked_apply_internal:
            referencer.apply_reference()
            mocked_apply_internal.assert_not_called()

    @patch("txrm2tiff.txrm.ref_mixin.file_can_be_opened", MagicMock(return_value=True))
    def test_apply_reference_tiff(self):
        referencer = ReferenceMixin()
        with patch.object(
            referencer, "apply_custom_reference_from_array"
        ) as mocked_apply_array:
            with TemporaryDirectory("tiff_ref_test_", dir=".") as tmp_dir:
                array = np.ones((3, 5, 5))
                tif_path = Path(tmp_dir) / "image.ome.tiff"
                manual_save(tif_path, array)

                referencer.apply_reference(tif_path, False)

        args = mocked_apply_array.call_args[0]
        kwargs = mocked_apply_array.call_args[1]
        assert_array_equal(args[0], array)
        self.assertEqual(kwargs, {"overwrite": True})

    @patch("txrm2tiff.txrm.ref_mixin.file_can_be_opened", MagicMock(return_value=True))
    def test_apply_reference_tiff_compensate_exposure(self):
        referencer = ReferenceMixin()
        with patch.object(
            referencer, "apply_custom_reference_from_array"
        ) as mocked_apply_array:
            with TemporaryDirectory("tiff_ref_test_", dir=".") as tmp_dir:
                array = np.ones((3, 5, 5))
                exposures = (1, 2, 3)
                metadata = OMEXML()
                metadata.image().Pixels.set_plane_count(len(exposures))
                for idx, exposure in enumerate(exposures):
                    metadata.image().Pixels.Plane(idx).set_ExposureTime(exposure)
                tif_path = Path(tmp_dir) / "image.ome.tiff"
                manual_save(tif_path, array, metadata=metadata)

                referencer.apply_reference(tif_path, True)
        args = mocked_apply_array.call_args[0]
        kwargs = mocked_apply_array.call_args[1]
        assert_array_equal(args[0], array)
        self.assertEqual(
            kwargs, {"overwrite": True, "custom_exposure": np.mean(exposures)}
        )

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    @patch("txrm2tiff.txrm.ref_mixin.file_can_be_opened", MagicMock(return_value=True))
    @patch("txrm2tiff.txrm.ref_mixin.isOleFile", MagicMock(return_value=True))
    @patch("txrm2tiff.txrm.ref_mixin.main.open_txrm")
    def test_apply_refererence_txrm(self, TxrmClass, mocked_open_txrm):
        referencer = ReferenceMixin()
        txrm = MagicMock(auto_spec=TxrmClass)
        mocked_open_txrm.return_value.__enter__.return_value = txrm

        with patch.object(referencer, "apply_reference_from_txrm") as mocked_from_txrm:
            referencer.apply_reference("test/path.txrm")
            mocked_from_txrm.assert_called_once_with(txrm)



import unittest
from unittest.mock import patch, MagicMock

from pathlib import Path
from datetime import datetime
from tempfile import NamedTemporaryFile, TemporaryDirectory
import numpy as np
from numpy.testing import assert_array_equal
from oxdls import OMEXML
import tifffile as tf

from txrm2tiff.utils.file_handler import (
    file_can_be_opened,
    ole_file_works,
    manual_save,
    manual_annotation_save,
)


class TestFileHandler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.fake_temp = NamedTemporaryFile(suffix=".txrm", delete=False)
        cls.fake_file = Path(cls.fake_temp.name)
        cls.real_file = Path(
            "/dls/science/groups/das/ExampleData/B24_test_data/data/2019/cm98765-1/raw/XMv10/test_tomo2_e3C_full.txrm"
        )

    @classmethod
    def tearDownClass(cls):
        cls.fake_temp.file.close()
        cls.fake_file.unlink()

    def test_ole_file_returns_true_for_real_file(self):
        if self.real_file.exists():
            self.assertTrue(ole_file_works(self.real_file))
        else:
            print("Cannot run test without access to dls directories")

    def test_real_file_can_be_opened(self):
        if self.real_file.exists():
            self.assertTrue(file_can_be_opened(self.real_file))
        else:
            print("Cannot run test without access to dls directories")

    def test_ole_file_returns_false_for_fake_file(self):
        self.assertFalse(ole_file_works(self.fake_file))

    def test_fake_file_can_be_opened(self):
        self.assertTrue(file_can_be_opened(self.fake_file))

    def test_nonexistent_file_can_be_opened(self):
        self.assertFalse(
            file_can_be_opened(
                Path(
                    "/fake/path/oh/no/{}".format(datetime.now().strftime("%Y%m%d_%H%M"))
                )
            )
        )

    def test_manual_save(self):
        with TemporaryDirectory(
            prefix="saving_test_", dir=Path(__name__).parent
        ) as tmpdir:
            im_path = Path(tmpdir) / "saved.tiff"
            image = np.zeros((5, 30, 30), dtype=np.float64)
            self.assertFalse(im_path.exists())
            manual_save(im_path, image)
            self.assertTrue(im_path.exists())

    def test_manual_save_with_datatype(self):
        target_dtype = np.uint16
        with TemporaryDirectory(
            prefix="saving_test_", dir=Path(__name__).parent
        ) as tmpdir:
            im_path = Path(tmpdir) / "saved.tiff"
            image = np.ones((5, 30, 35), dtype=np.float64)
            self.assertFalse(im_path.exists())
            manual_save(im_path, image, data_type=target_dtype)
            self.assertTrue(im_path.exists())
            with tf.TiffFile(im_path) as tiff:
                saved_arr = tiff.asarray()

        self.assertEqual(saved_arr.dtype, target_dtype)
        assert_array_equal(saved_arr, image)

    def test_manual_save_with_metadata(self):
        target_dtype = np.uint16

        with TemporaryDirectory(
            prefix="saving_test_", dir=Path(__name__).parent
        ) as tmpdir:
            im_path = Path(tmpdir) / "saved.tiff"
            metadata = OMEXML()

            image = np.ones((5, 30, 35), dtype=np.float64)
            self.assertFalse(im_path.exists())
            manual_save(im_path, image, data_type=np.uint16, metadata=metadata)
            self.assertTrue(im_path.exists())
            with tf.TiffFile(im_path) as tiff:
                saved_arr = tiff.asarray()
                saved_meta = tiff.pages[0].description

        # Image name and data type should be set while being saved
        metadata.image().set_Name(im_path.name)
        metadata.image().Pixels.set_PixelType("uint16")

        self.assertEqual(saved_arr.dtype, target_dtype)
        assert_array_equal(saved_arr, image)
        self.assertEqual(saved_meta, metadata.to_xml().strip())  # Remove final newline

    def test_manual_save_sets_pixel_size(self):
        with TemporaryDirectory(
            prefix="saving_test_", dir=Path(__name__).parent
        ) as tmpdir:
            im_path = Path(tmpdir) / "saved.tiff"
            metadata = OMEXML()
            pixel_size_xy = (1, 2)  # nm
            pixels = metadata.image().Pixels
            pixels.set_PhysicalSizeX(pixel_size_xy[0])
            pixels.set_PhysicalSizeY(pixel_size_xy[1])
            image = np.ones((5, 30, 35), dtype=np.float64)
            self.assertFalse(im_path.exists())
            manual_save(im_path, image, data_type=np.uint16, metadata=metadata)
            self.assertTrue(im_path.exists())
            with tf.TiffFile(im_path) as tiff:
                saved_arr = tiff.asarray()
                x_resolution = tiff.pages[0].tags['XResolution'].value
                y_resolution = tiff.pages[0].tags['YResolution'].value
                resolution_unit = tiff.pages[0].tags['ResolutionUnit'].value

        assert_array_equal(saved_arr, image)
        self.assertEqual(x_resolution, (int(1.e7), 1))
        self.assertEqual(y_resolution, (int(5.e6), 1))
        self.assertEqual(resolution_unit, int(tf.TIFF.RESUNIT.CENTIMETER))


    def test_manual_annotation_save(self):
        with TemporaryDirectory(
            prefix="annotation_saving_test_", dir=Path(__name__).parent
        ) as tmpdir:
            im_path = Path(tmpdir) / "saved.tiff"
            image = np.ones((5, 30, 35, 3))
            self.assertFalse(im_path.exists())
            manual_annotation_save(im_path, image)
            self.assertTrue(im_path.exists())
            with tf.TiffFile(im_path) as tiff:
                saved_arr = tiff.asarray()

        assert_array_equal(saved_arr, image)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock

from txrm2tiff.txrm.txrm_property import txrm_property


class ToasterError(Exception):
    pass


class Toaster:
    def __init__(self):
        self.ole = None
        self.strict = False
        self.file_is_open = True

    @txrm_property(fallback="bread")
    def toast(self):
        return self.heater()

    def heater(self):
        return "nice toast"


class TestTxrmProperty(unittest.TestCase):
    def test_txrm_property_fails_as_method(self):
        t = Toaster()
        with self.assertRaises(TypeError):
            t.toast()
            # Not a method

    def test_txrm_property_works_as_property(self):
        t = Toaster()
        self.assertEqual(t.toast, "nice toast")

    def test_delete_works(self):
        t = Toaster()
        t.heater = MagicMock(side_effect=["nice toast", "burnt toast"])
        # Check that toast's return value is cached until deleted
        self.assertEqual(t.toast, "nice toast")
        self.assertEqual(t.toast, "nice toast")
        self.assertEqual(t.toast, "nice toast")
        del t.toast
        self.assertIs(t.toast, "burnt toast")

    def test_error_returns_fallback(self):
        t = Toaster()
        t.file_is_open = True
        t.heater = MagicMock(side_effect=ToasterError("The toaster is not plugged in."))
        self.assertEqual(t.toast, "bread")

    def test_error_and_unopened_file_returns_fallback(self):
        t = Toaster()
        t.file_is_open = False
        t.heater = MagicMock(side_effect=ToasterError("The toaster is not plugged in."))
        self.assertEqual(t.toast, "bread")

    def test_strict_error_and_unopened_raises_IOError(self):
        t = Toaster()
        t.strict = True
        t.file_is_open = False
        t.heater = MagicMock(side_effect=ToasterError("The toaster is not plugged in."))
        with self.assertRaises(IOError):
            t.toast

    def test_strict_error_and_opened_raises_function_error(self):
        t = Toaster()
        t.strict = True
        t.file_is_open = True
        t.heater = MagicMock(side_effect=ToasterError("The toaster is not plugged in."))
        with self.assertRaises(ToasterError):
            t.toast

    def test_error_raised_if_strict(self):
        t = Toaster()
        t.strict = True
        t.heater = MagicMock(side_effect=ToasterError("The toaster is not plugged in."))
        with self.assertRaises(ToasterError):
            t.toast

    def test_cannot_set_attribute(self):
        t = Toaster()
        self.assertEqual(t.toast, "nice toast")
        with self.assertRaises(AttributeError):
            t.toast = "bread"
            # You cannot un-toast



import traceback
from dataclasses import dataclass

from txrm2tiff.xradia_properties import stream_dtypes

from .txrm.abstract import AbstractTxrm
from .txrm_functions import general
from .xradia_properties import XrmDataTypes


class Inspector:

    def __init__(self, txrm:AbstractTxrm):
        self.txrm = txrm
        self._output_text = ""

    def inspect(self, extra=False):
        self._output_text += f"\n{self.txrm.name}\t(v{self.txrm.version} file)\n"

        self._output_text += "------------------------------------------------------------------------------------------\n"

        self.inspect_basic()
        self._output_text += "------------------------------------------------------------------------------------------\n\n"

        if extra:
            self.inspect_extra()
            self._output_text += "------------------------------------------------------------------------------------------\n\n"

    def inspect_basic(self):
        images_taken = self.txrm.image_info.get("ImagesTaken", (0,))[0]
        num_images = self.txrm.image_info.get("NoOfImages", (0,))[0]
        if images_taken != num_images:
            num_str = f"{images_taken} images (of {num_images} planned)"
        else:
            num_str = f"{images_taken} images"
        self._output_text += "{0} of type {1} with dimensions: {2}\n".format(
            num_str,
            self.txrm.image_dtype
            if self.txrm.image_dtype is None
            else self.txrm.image_dtype,
            ", ".join([str(i) for i in self.txrm.image_dims]),
        )
        if self.txrm.shape[::-1] != self.txrm.image_dims:
            # Currently only used for v3 mosaics but may be useful if analysing a processed txrm object
            self._output_text += f"The images are stored as an array of shape (rows x columns): {self.txrm.shape[0]}x{self.txrm.shape[1]}\n"

        if self.txrm.is_mosaic:
            self._output_text += f"Is a mosaic of shape (rows x coumns): {self.txrm.mosaic_dims[0]}x{self.txrm.mosaic_dims[1]}\n"
        else:
            self._output_text += "Not a mosaic\n"

        self._output_text += "Pixel size: {0}μm\n".format(
            self.txrm.image_info.get("PixelSize", (0,))[0]
        )

        if self.txrm.has_stream("ReferenceData/Image"):
            self._output_text += (
                "Reference of type {0} applied with dimensions: {1}\n".format(
                    self.txrm.reference_dtype,
                    ", ".join([str(i) for i in self.txrm.reference_dims]),
                )
            )
        else:
            self._output_text += "No reference applied\n"

    def inspect_extra(self):
        self._inspect_image_info()
        self._output_text += "\n"
        self._inspect_reference_info()
        self._output_text += "\n"
        self._inspect_position_info()

    def _inspect_image_info(self):
        if self.txrm.image_info:
            self._output_text += "ImageInfo streams:\n"
            for name, values in self.txrm.image_info.items():
                self._output_text += "\t{0}: {1}\n\n".format(
                    name, ", ".join([f"{p}" for p in values])
                )

    def _inspect_reference_info(self):
        if self.txrm.reference_info:
            self._output_text += "ReferenceData/ImageInfo streams:\n"
            for name, values in self.txrm.reference_info.items():
                self._output_text += "\t{0}: {1}\n\n".format(
                    name, ", ".join([f"{p}" for p in values])
                )

    def _inspect_position_info(self):
        if self.txrm.position_info:
            self._output_text += "PositionInfo streams:\n"
            for name, (pos, unit) in self.txrm.position_info.items():
                self._output_text += "\t{0} ({1}): {2}\n\n".format(
                    name, unit, ", ".join([f"{p:.3f}" for p in pos])
                )
            # Displays values to 3dp in Data Explorer

    def _get_stream_list(self):
        return self.txrm.list_streams()

    def list_streams(self):
        for stream in self._get_stream_list():
            stream_size = self.txrm.ole.get_size(stream)
            self._output_text += f"{stream:110s} Size: {stream_size:6d} bytes\n"

    def inspect_streams(self, *keys):
        self._output_text += "Inspecting streams: " + ", ".join(keys) + "\n"
        try:
            for key in keys:
                if self.txrm.has_stream(key):

                    self._output_text += f"\n\n{key}:"
                    if key in stream_dtypes.streams_dict:
                        try:
                            dtype = stream_dtypes.streams_dict.get(key)
                            values = general.read_stream(
                                self.txrm.ole,
                                key,
                                dtype,
                                strict=True,
                            )
                            values_str = ", ".join([str(i) for i in values])
                            self._output_text += f"\t{values_str} (stored as {dtype})"
                        except (ValueError, TypeError) as e:
                            self._output_text += (
                                f"\tWARNING: Expected data type unsuccessful:\t{e}\n"
                            )
                            self.inspect_unknown_dtype_stream(key)
                    else:
                        self._output_text += f"\nUnknown data type. "
                        self.inspect_unknown_dtype_stream(key)
                else:
                    self._output_text += f"\nStream '{key}' does not exist.\n"
        except Exception as e:
            self._output_text += f"Unexpected exception reading stream {key}:\n\n{''.join(traceback.format_exception(type(e), e, e.__traceback__))}\n\n"

    def get_text(self):
        return self._output_text

    def inspect_unknown_dtype_stream(self, key: str) -> None:
        self._output_text += "Trying all XRM data types:\n"
        for dtype_enum in XrmDataTypes:
            try:
                values_str = ", ".join(
                    [
                        str(i)
                        for i in general.read_stream(
                            self.txrm.ole, key, dtype_enum, strict=True
                        )
                    ]
                )
                self._output_text += f"\t{dtype_enum.name + ':':18s}\t{values_str}\n\n"
            except (ValueError, TypeError) as e:
                self._output_text += f"\t{dtype_enum.name + ':':18s}\t{e}\n\n"



import unittest
from unittest.mock import patch, MagicMock
from numpy.testing import assert_array_equal

import numpy as np

from txrm2tiff.txrm.v5 import Txrm5


@patch.multiple(Txrm5, open=MagicMock)
class TestTxrm5(unittest.TestCase):
    @patch("txrm2tiff.txrm.v5.AbstractTxrm.mosaic_dims", new=(2, 3))
    def test_is_mosaic_true(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertTrue(txrm.is_mosaic)

    @patch("txrm2tiff.txrm.v5.AbstractTxrm.mosaic_dims", new=(0, 0))
    def test_is_mosaic_false(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertFalse(txrm.is_mosaic)

    def test_is_mosaic_false_if_no_mode(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertFalse(txrm.is_mosaic)

    @patch(
        "txrm2tiff.txrm.v5.AbstractTxrm.image_info",
        new={"ImageWidth": [20], "ImageHeight": [5]},
    )
    def test_image_dims(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertEqual(txrm.image_dims, [20, 5])

    @patch(
        "txrm2tiff.txrm.v5.AbstractTxrm.reference_info",
        new={"ImageWidth": [20], "ImageHeight": [5]},
    )
    def test_ref_image_dims(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertEqual(txrm.reference_dims, [20, 5])

    @patch("txrm2tiff.txrm.v5.AbstractTxrm.reference_info", new={"ExpTime": [0.5]})
    def test_reference_exposure(self):
        txrm = Txrm5("test/path/file.txrm")
        self.assertEqual(txrm.reference_exposure, 0.5)

    @patch("txrm2tiff.txrm.v5.Txrm5.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v5.Txrm5.reference_dims", new=[2, 3])
    @patch("txrm2tiff.txrm.v5.AbstractTxrm.extract_reference_data")
    @patch("txrm2tiff.txrm.v5.fallback_image_interpreter")
    def test_extract_reference_image_unknown_dtype(
        self, mocked_fallback_interpreter, mocked_extract_ref_data
    ):
        strict = False
        txrm = Txrm5(
            "test/path/file.txrm",
            load_images=False,
            load_reference=False,
            strict=strict,
        )
        ole = MagicMock()
        txrm.ole = ole
        im_size = np.prod(txrm.reference_dims)
        arr = np.arange(im_size, dtype=np.uint16)
        arr.shape = txrm.reference_dims[::-1]
        arr_flat = arr.flatten()
        arr_bytes = arr_flat.tobytes()
        mocked_extract_ref_data.return_value = arr_bytes
        mocked_fallback_interpreter.return_value = arr_flat

        output = txrm.extract_reference_image()

        mocked_extract_ref_data.assert_called_once_with()
        mocked_fallback_interpreter.assert_called_once_with(arr_bytes, im_size, strict)

        assert_array_equal(output, arr)

    @patch("txrm2tiff.txrm.v5.Txrm5.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v5.Txrm5.reference_dims", new=[2, 3])
    @patch("txrm2tiff.txrm.v5.AbstractTxrm.extract_reference_data")
    @patch("txrm2tiff.txrm.v5.fallback_image_interpreter")
    def test_extract_reference_image_known_dtype(
        self, mocked_fallback_interpreter, mocked_extract_ref_data
    ):
        strict = False
        txrm = Txrm5(
            "test/path/file.txrm",
            load_images=False,
            load_reference=False,
            strict=strict,
        )
        ole = MagicMock()
        txrm.ole = ole
        im_size = np.prod(txrm.reference_dims)
        arr = np.arange(im_size, dtype=np.uint16)
        arr.shape = txrm.reference_dims[::-1]
        arr_flat = arr.flatten()
        mocked_extract_ref_data.return_value = arr_flat

        output = txrm.extract_reference_image()

        mocked_extract_ref_data.assert_called_once_with()
        mocked_fallback_interpreter.assert_not_called()

        assert_array_equal(output, arr)



from os import PathLike
import typing
import logging
from numbers import Number
import numpy as np
import tifffile as tf
from olefile import isOleFile
from oxdls import OMEXML
from tifffile.tifffile import TiffFileError

from . import main
from .abstract import AbstractTxrm
from ..utils.file_handler import file_can_be_opened
from ..utils.image_processing import dynamic_despeckle_and_average_series
from ..utils.functions import conditional_replace


class ReferenceMixin:
    def apply_reference(
        self,
        custom_reference: typing.Optional[typing.Union[str, PathLike]] = None,
        compensate_exposure: bool = True,
        overwrite: bool = True,
    ) -> None:
        if custom_reference is not None and file_can_be_opened(custom_reference):
            if isOleFile(str(custom_reference)):
                with main.open_txrm(custom_reference) as ref_txrm:
                    self.apply_reference_from_txrm(ref_txrm)
            else:
                return self._apply_reference_from_tiff(
                    custom_reference, compensate_exposure, overwrite
                )
        elif self.has_reference:
            return self.apply_internal_reference(
                overwrite=overwrite, compensate_exposure=compensate_exposure
            )
        else:
            logging.info("No reference to apply")

    def _apply_reference_from_tiff(
        self,
        custom_reference: PathLike,
        compensate_exposure: bool = True,
        overwrite: bool = True,
    ):
        try:
            with tf.TiffFile(str(custom_reference)) as tif:
                kwargs = {}
                if compensate_exposure:
                    try:
                        pixels = OMEXML(xml=tif.pages[0].description).image().Pixels
                        plane_count = pixels.get_plane_count()
                        kwargs["custom_exposure"] = np.mean(
                            [
                                pixels.Plane(i).get_ExposureTime()
                                for i in range(plane_count)
                            ]
                        )
                    except Exception:
                        logging.warning(
                            "Unable to extract exposure(s) from TIFF metadata - reference will not be scaled.",
                            exc_info=True,
                        )
                arr = tif.asarray()
                return self.apply_custom_reference_from_array(
                    arr, overwrite=overwrite, **kwargs
                )
        except TiffFileError:
            logging.error(
                "Invalid custom reference supplied: file is not a TXRM or TIFF file."
            )
            return

    def apply_custom_reference_from_array(
        self,
        custom_reference: np.ndarray,
        custom_exposure: typing.Optional[Number] = None,
        overwrite: bool = True,
    ) -> np.ndarray:
        """Applies numpy array as a reference image to txrm image(s), returning the images as a numpy ndarray and overwrites the txrm image(s) if overwrite is True."""
        if self.referenced:
            logging.warning(
                "Applying reference to already referenced txrm image(s). If you did not mean to reference more than once, reload the images before applying the correct reference."
            )
        try:
            ref_img = ReferenceMixin._flatten_reference(custom_reference)
            ref_img = self._tile_reference_if_needed(custom_reference)
        except Exception:
            if self.strict:
                raise
            logging.error(
                "Exception occurred preparing the reference for %s",
                self.name,
                exc_info=True,
            )
        if custom_exposure is not None:
            self._compensate_ref_exposure(ref_img, custom_exposure)
        ref = ReferenceMixin._apply_reference_to_images(self.get_images(), ref_img)
        if overwrite:
            self._images = ref
            self.referenced = True
        return ref

    def apply_reference_from_txrm(
        self,
        txrm: AbstractTxrm,
        compensate_exposure: bool = True,
        overwrite: bool = True,
    ) -> np.ndarray:
        """Applies image(s) to txrm image(s), returning the images as a numpy ndarray and overwrites the txrm image(s) if overwrite is True."""
        self.apply_custom_reference_from_array(
            txrm.get_images(load=True),
            np.mean(txrm.exposures),
            compensate_exposure,
            overwrite,
        )

    def apply_internal_reference(
        self, compensate_exposure: bool = True, overwrite: bool = True
    ) -> np.ndarray:
        """Applies reference to txrm image(s), returning the images as a numpy ndarray and overwrites the txrm image(s) if overwrite is True."""
        if self.referenced:
            logging.warning(
                "Applying reference to already referenced txrm image(s). If you did not mean to reference more than once, reload the images before applying the correct reference."
            )
        ref_img = self.get_reference(load=True)
        logging.info("Internal reference will be applied to %s", self.name)
        if ref_img is None:
            if self.strict:
                raise AttributeError("No internal reference to apply")
            logging.warning("No internal reference to apply")
            return
        if compensate_exposure:
            ref_img = self._compensate_ref_exposure(ref_img, self.reference_exposure)
        referenced = ReferenceMixin._apply_reference_to_images(
            self.get_images(load=True), ref_img
        )
        if overwrite:
            self._images = referenced
            self.referenced = True
        return referenced

    def _tile_reference_if_needed(self, custom_reference: np.ndarray) -> np.ndarray:
        """Tile the image if it is needed to match a mosaic Assumes axes [y, x]."""
        needs_stitching = self.is_mosaic and custom_reference.shape == [
            round(img_dim / mos_dim, 2)
            for img_dim, mos_dim in zip(self.shape, self.mosaic_dims[::-1])
        ]  # True if it's a mosaic and the correct dims to be tiled to mosaic
        assert custom_reference.shape == self.shape or needs_stitching, (
            "Invalid reference shape for %s" % self.name
        )
        # Checks that reference is either the size of the image or can be stitched to that size
        if needs_stitching:
            return np.tile(
                custom_reference, self.mosaic_dims[::-1]
            )  # Tiles reference if needed
        return custom_reference

    def _compensate_ref_exposure(self, ref_image, ref_exposure):
        # TODO: Improve this in line with ALBA's methodolgy
        # Normalises the reference exposure
        # Assumes roughly linear response, which is a reasonable estimation
        # because exposure times are unlikely to be significantly different)
        # (if it is a tomo, it does this relative to the 0 degree image, not on a per-image basis)
        multiplier = self.exposures[self.zero_angle_index] / ref_exposure
        return ref_image.copy() * multiplier

    @staticmethod
    def _apply_reference_to_images(
        images: np.ndarray, reference: np.ndarray
    ) -> np.ndarray:
        floated_and_referenced = (
            np.asarray(images, dtype=np.float32) * 100.0 / reference
        )
        if (
            np.isnan(floated_and_referenced).any()
            or np.isinf(floated_and_referenced).any()
        ):
            logging.warning(
                "Potential dead pixels found. "
                "NaN was output for at least one pixel in the referenced image."
            )
            # Replace any infinite pixels (nan or inf) with 0:
            conditional_replace(floated_and_referenced, 0, lambda x: ~np.isfinite(x))
        return floated_and_referenced.astype(np.float32)

    @staticmethod
    def _flatten_reference(image: np.ndarray) -> np.ndarray:
        """
        Despeckle and average images if they are an image stack. Assumes axes [idx, y, x] and tile the image if needed.

        Returns numpy array with axes [y, x] (flattens axes of size 1)
        """
        if len(image.shape) == 3:
            if image.shape[0] > 1:
                image = dynamic_despeckle_and_average_series(image, average=True)
            else:
                image = np.squeeze(image)
        return image



import logging
import math
from numbers import Number
from pathlib import Path
from collections import defaultdict
from typing import Iterable, Optional, Tuple

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageOps import autocontrast

from ..utils.image_processing import rescale_image
from ..utils.file_handler import manual_annotation_save
from ..xradia_properties import AnnotationTypes, XrmDataTypes as XDT

font_path = Path(__file__).parent.parent / "font" / "CallingCode-Regular.otf"


class AnnotatorMixin:
    def __new__(cls, *args, **kwargs):
        cls.setup_ann_functions()
        return super().__new__(cls)

    @classmethod
    def setup_ann_functions(cls):
        cls.ann_funcs = defaultdict(
            lambda: cls._not_implemented,
            {
                AnnotationTypes.ANN_LINE: cls._plot_line,
                AnnotationTypes.ANN_RECT: cls._plot_rect,
                AnnotationTypes.ANN_ELLIPSE: cls._plot_ellipse,
                AnnotationTypes.ANN_CIRCLE: cls._plot_ellipse,
                AnnotationTypes.ANN_POLYGON: cls._plot_polygon,
                AnnotationTypes.ANN_POLYLINE: cls._plot_polyline,
                AnnotationTypes.ANN_FREE_HAND_SKETCH: cls._plot_freehand,
            },
        )

    def annotate(
        self,
        scale_bar: bool = True,
        clip_percentiles: Tuple[int | float, int | float] = (2, 98),
    ) -> Optional[np.ndarray]:
        """Annotate output image. Please ensure that the image has been referenced first, if applicable."""
        annotations = self.extract_annotations(scale_bar=scale_bar)
        # Checks if anything has been added
        if annotations is None:
            logging.warning("No annotations to apply")
            self.annotated_image = None
        else:
            # Annotations will be in the wrong place if flipped
            images = self.get_output(flip=False, clear_images=False)
            lower_percentile, upper_percentile = np.percentile(images, clip_percentiles)
            images = rescale_image(
                images,
                0,
                255,
                previous_minimum=lower_percentile,
                previous_maximum=upper_percentile,
            )
            self.annotated_image = self.apply_annotations(images, annotations)
        return self.annotated_image

    @property
    def thickness_modifier(self) -> float:
        """This set a multiplier so that lines are visible when the image is at a sensible size"""
        if not hasattr(self, "output_shape"):
            return 1.0
        return np.mean(self.output_shape[1:]) / 500.0

    def _create_image_and_draw(self):
        """Create transparent RGBA image with appropriate 2D dimensions for output image(s)"""
        im = Image.new("RGBA", self.output_shape[:0:-1], (0, 0, 0, 0))
        draw = ImageDraw.Draw(im, mode="RGBA")
        return im, draw

    def extract_annotations(self, scale_bar: bool = True) -> Optional[np.ndarray]:
        annotations, draw = self._create_image_and_draw()
        annotated = False
        if self.has_stream("Annot/TotalAnn"):
            num_annotations = self.read_stream("Annot/TotalAnn", strict=True)[0]
            if num_annotations is not None and num_annotations > 0:
                for i in range(num_annotations):
                    try:
                        self._draw_annotation(draw, i)
                        annotated = True
                    except Exception:
                        # Create error message but don't stop other annotations from being extracted
                        logging.error("Failed to get annotation %i", i, exc_info=True)
        # _add_scale_bar() returns True if scale bar successfully added
        if scale_bar:
            annotated = self._add_scale_bar(draw) or annotated
        if not annotated:
            logging.warning("No annotations were extracted")
            return None
        return annotations

    def apply_annotations(self, images: np.ndarray, annotations: Image) -> np.ndarray:
        """
        Apply annotations to images returning a numpy.ndarray

        Args:
            images (3D NumPy array): Images that will be annotated

        Returns:
            RGB numpy array or None: Annotated images, if successfully applied
        """
        try:
            # Create output array with shape of input, plus 3 channels for RGB
            output_images = np.zeros((*images.shape, 3), dtype="uint8")
            for idx, image in enumerate(images, 0):
                # Make 'L' type PIL image from 2D array, autocontrast, then convert to RGBA
                image = Image.fromarray(image).convert("L")
                image_auto_contrasted = autocontrast(image)
                if np.mean(image_auto_contrasted) == np.max(image_auto_contrasted):
                    image = image_auto_contrasted
                else:
                    del image_auto_contrasted
                image = image.convert("RGBA")
                # Combine annotations and image by alpha
                image.alpha_composite(annotations)
                # Throw away alpha and fill into output array
                output_images[idx] = image.convert("RGB")
            return output_images
        except Exception:
            logging.error("Failed to apply annotations", exc_info=True)

    def _draw_annotation(self, draw: ImageDraw, index: int) -> None:
        stream_stem = "Annot/Ann%i" % index
        ann_type_int = self.read_stream(
            f"{stream_stem}/AnnType", XDT.XRM_INT, strict=True
        )[0]
        self.ann_funcs.get(AnnotationTypes(ann_type_int))(self, draw, stream_stem)

    def _add_scale_bar(self, draw: ImageDraw) -> bool:
        pixel_size = self.image_info["PixelSize"][0]  # microns
        if pixel_size is not None and pixel_size > 0:
            try:
                colour = (0, 255, 0, 255)  # Solid green
                # Set up scale bar dims:
                x0, y0 = self._flip_y(
                    (self.output_shape[2] // 50, self.output_shape[1] // 30)
                )[0]
                bar_width = int(6 * self.thickness_modifier)
                # Set up scale bar text:
                f = ImageFont.truetype(
                    str(font_path), int(15 * self.thickness_modifier)
                )

                tmp_bar_length = (
                    self.output_shape[2] / 5.0
                )  # Get initial bar length based on image width
                tmp_bar_size = tmp_bar_length * pixel_size  # Calculate physical size

                # Round to nearest multiple of the order of magnitude
                exponent = math.floor(math.log10(tmp_bar_size))
                step_size = 10.0**exponent
                bar_size = round(
                    round(tmp_bar_size / step_size)
                    * step_size,  # Find nearest multiple of step_size
                    -exponent,
                )  # round to -exponent decimal places (to mitigate rounding errors)
                bar_length = (
                    bar_size / pixel_size
                )  # Calculate number of pixels that wide new bar size bar should be

                # Set text and calculate text positions:
                text = "%gμm" % bar_size
                text_xy = (x0 + bar_length / 2, y0 - bar_width)
                # Draw:
                draw.text(
                    text_xy, text, font=f, fill=colour, anchor="mb"
                )  # anchor: middle-bottom
                draw.line((x0, y0, x0 + bar_length, y0), fill=colour, width=bar_width)
                return True
            except Exception:
                logging.error(
                    "Exception occurred while drawing scale bar", exc_info=True
                )
        return False

    def _get_thickness(self, stream_stem: str) -> int:
        return int(
            self.read_stream(
                f"{stream_stem}/AnnStrokeThickness", XDT.XRM_DOUBLE, strict=True
            )[0]
            * self.thickness_modifier
        )

    def _get_colour(self, stream_stem: str) -> Tuple:
        colours = self.read_stream(f"{stream_stem}/AnnColor", XDT.XRM_UNSIGNED_CHAR)
        return (*colours[2::-1], colours[3])

    @staticmethod
    def _not_implemented(*args):
        """Do nothing"""
        pass

    def _plot_line(self, draw: ImageDraw, stream_stem: str) -> None:
        colour = self._get_colour(stream_stem)
        thickness = self._get_thickness(stream_stem)
        x1 = self.read_stream(f"{stream_stem}/X1", XDT.XRM_FLOAT, strict=True)[0]
        x2 = self.read_stream(f"{stream_stem}/X2", XDT.XRM_FLOAT, strict=True)[0]
        y1 = self.read_stream(f"{stream_stem}/Y1", XDT.XRM_FLOAT, strict=True)[0]
        y2 = self.read_stream(f"{stream_stem}/Y2", XDT.XRM_FLOAT, strict=True)[0]

        draw.line(self._flip_y((x1, y1), (x2, y2)), fill=colour, width=thickness)

    def __extract_rectangle_coords(self, stream_stem: str) -> Tuple[Tuple[int]]:
        x = []
        y = []
        x.append(
            self.read_stream(f"{stream_stem}/Rectangle/Left", XDT.XRM_INT, strict=True)[
                0
            ]
        )
        y.append(
            self.read_stream(
                f"{stream_stem}/Rectangle/Bottom", XDT.XRM_INT, strict=True
            )[0]
        )
        x.append(
            self.read_stream(
                f"{stream_stem}/Rectangle/Right", XDT.XRM_INT, strict=True
            )[0]
        )
        y.append(
            self.read_stream(f"{stream_stem}/Rectangle/Top", XDT.XRM_INT, strict=True)[
                0
            ]
        )
        x.sort()
        y.sort(reverse=True)
        # Values in the first tuple must be smaller than their respective second tuple value.
        # Therefore y1 must be first and y0 second because they are flipped.
        return self._flip_y(*zip(x, y))

    def _plot_rect(self, draw: ImageDraw, stream_stem: str) -> None:
        colour = self._get_colour(stream_stem)
        thickness = self._get_thickness(stream_stem)
        xy = self.__extract_rectangle_coords(stream_stem)

        draw.rectangle(xy, outline=colour, width=thickness)

    def _plot_ellipse(self, draw: ImageDraw, stream_stem: str) -> None:
        colour = self._get_colour(stream_stem)
        thickness = self._get_thickness(stream_stem)
        xy = self.__extract_rectangle_coords(stream_stem)

        draw.ellipse(xy, outline=colour, width=thickness)

    def _plot_polygon(self, draw: ImageDraw, stream_stem: str) -> None:
        colour = self._get_colour(stream_stem)
        thickness = self._get_thickness(stream_stem)
        total_points = self.read_stream(
            f"{stream_stem}/TotalPts", XDT.XRM_UNSIGNED_INT
        )[0]
        xs = self.read_stream(f"{stream_stem}/PointX", XDT.XRM_FLOAT, strict=True)[
            :total_points
        ]
        ys = self.read_stream(f"{stream_stem}/PointY", XDT.XRM_FLOAT, strict=True)[
            :total_points
        ]
        xs.append(xs[0])  # Link beginning and end points
        ys.append(ys[0])

        draw.line(self._flip_y(*zip(xs, ys)), fill=colour, width=thickness)

    def _plot_polyline(
        self, draw: ImageDraw, stream_stem: str, joint: Optional[str] = None
    ):
        colour = self._get_colour(stream_stem)
        thickness = self._get_thickness(stream_stem)
        total_points = self.read_stream(
            f"{stream_stem}/TotalPts", XDT.XRM_UNSIGNED_INT
        )[0]
        xs = self.read_stream(f"{stream_stem}/PointX", XDT.XRM_FLOAT, strict=True)
        ys = self.read_stream(f"{stream_stem}/PointY", XDT.XRM_FLOAT, strict=True)

        draw.line(
            self._flip_y(*zip(xs[:total_points], ys[:total_points])),
            fill=colour,
            width=thickness,
            joint=joint,
        )

    def _plot_freehand(self, draw: ImageDraw, stream_stem: str) -> None:
        self._plot_polyline(draw, stream_stem, joint="curve")

    def _flip_y(self, *xys: Iterable[Number]) -> Tuple[Tuple[Number]]:
        """
        Stored x-y coordinates are assuming 0 is bottom left, whereas PIL assumes 0 is top left
        This flips any list of coordinates that alternate x, y

        Args:
            *xys: n args assuming the form (x0, y0), (x1, y1), ... (xn, yn)

        Returns:
            Tuple of tuples: x-y coordinates with a flipped y
        """
        return tuple((x, self.output_shape[1] - 1 - y) for x, y in xys)

    def save_annotations(
        self,
        filepath: Optional[Path] = None,
        mkdir: bool = False,
        strict: Optional[bool] = None,
    ) -> bool:
        """Saves images (if available) returning True if successful."""
        if strict is None:
            strict = self.strict
        try:
            if self.annotated_image is None:
                raise AttributeError("No annotated image to save")
            if filepath is None:
                if self.path is None:
                    raise ValueError(
                        "An output filepath must be given if an input path was not given."
                    )
                filepath = self.path.resolve()
                filepath = filepath.parent / f"{filepath.stem}_Annotated.tiff"

            if mkdir:
                filepath.parent.mkdir(parents=True, exist_ok=True)

            manual_annotation_save(
                filepath,
                self.annotated_image,
            )
            return True
        except Exception:
            logging.error("Saving failed", exc_info=not strict)
            if strict:
                raise
            return False



import unittest
from parameterized import parameterized
from unittest.mock import MagicMock, patch
from numpy.testing import assert_array_equal

import struct
import numpy as np

from txrm2tiff import txrm_functions
from txrm2tiff.xradia_properties.enums import XrmDataTypes


class TestTxrmFunctions(unittest.TestCase):
    def test_reading_string_stream(self):
        ole = MagicMock()
        stream = MagicMock()
        stream_name = "Test"
        ole.openstream.return_value = stream
        ole.exists.return_value = True
        stream.read.return_value = (
            b"Pixis\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        )

        stream_list = txrm_functions.general._read_text_stream_to_list(ole, stream_name)

        ole.openstream.assert_called_with(stream_name)

        self.assertEqual(stream_list, ["Pixis"], msg="output is not as expected")

    @parameterized.expand(
        [(x,) for x in XrmDataTypes if x != XrmDataTypes.XRM_STRING],
    )  # Strings are handled by _read_text_stream_to_list
    def test_read_XRM_data_types_from_stream(self, dtype):
        ole = MagicMock()
        stream = MagicMock()
        stream_name = f"Test_{dtype.name}"
        ole.openstream.return_value = stream
        ole.exists.return_value = True
        values = np.arange(0, 100).astype(dtype.value)
        stream.getvalue.return_value = values.tobytes()
        try:
            output = txrm_functions.general._read_number_stream_to_list(
                ole, stream_name, dtype=dtype.value
            )
        except Exception:
            print(f"Failed for {dtype}")
            raise
        self.assertEqual(output, values.tolist())

    @parameterized.expand(
        [(x,) for x in XrmDataTypes if x != XrmDataTypes.XRM_STRING],
    )  # Images are not made of strings
    @patch("txrm2tiff.txrm_functions.images.extract_image_dtype")
    def test_extracts_image_with_XRM_data_types(self, dtype, patched_dtype_extractor):
        im_shape = (6, 9)

        ole = MagicMock()
        ole.exists.return_value = True

        stream = MagicMock()
        ole.openstream.return_value = stream
        image = np.arange(np.prod(im_shape), dtype=dtype.value)
        image.shape = im_shape
        stream.getvalue.return_value = image.tobytes()
        patched_dtype_extractor.return_value = dtype
        output = txrm_functions.extract_single_image(
            ole, 24, im_shape[0], im_shape[1], strict=True
        )

        assert_array_equal(
            image, output, err_msg=f"Failed to unpack image with dtype={dtype}"
        )

    @parameterized.expand(
        [(XrmDataTypes.XRM_UNSIGNED_SHORT,), (XrmDataTypes.XRM_FLOAT,)]
    )
    @patch("txrm2tiff.txrm_functions.images.extract_image_dtype")
    def test_extracting_single_image_unknown_type(self, dtype, patched_dtype_extractor):
        im_shape = (6, 9)

        ole = MagicMock()
        ole.exists.return_value = True

        stream = MagicMock()
        ole.openstream.return_value = stream
        image = np.arange(np.prod(im_shape), dtype=dtype.value)
        image.shape = im_shape
        stream.getvalue.return_value = image.tobytes()
        patched_dtype_extractor.return_value = None
        output = txrm_functions.extract_single_image(
            ole, 24, im_shape[0], im_shape[1], strict=True
        )

        assert_array_equal(
            image, output, err_msg=f"Failed to unpack image with dtype={dtype}"
        )

    def test_extracting_single_image_short(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        ole.exists.side_effect = [True, False]
        stream.getvalue.return_value = struct.pack("<6H", *list(range(6)))

        data = txrm_functions.extract_single_image(ole, 1, 2, 3)

        ole.openstream.assert_called_with("ImageData1/Image1")

        assert_array_equal(
            data, np.arange(6).reshape(2, 3), err_msg="output is not as expected"
        )

    def test_extracting_single_image_float(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        ole.exists.side_effect = [True, False]
        stream.getvalue.return_value = struct.pack("<6f", *list(range(6)))

        data = txrm_functions.extract_single_image(ole, 1, 2, 3)

        ole.openstream.assert_called_with("ImageData1/Image1")

        assert_array_equal(
            data, np.arange(6).reshape(2, 3), err_msg="output is not as expected"
        )

    def test_read_stream_failure(self):
        ole = MagicMock()
        ole.exists.return_value = False
        data = txrm_functions.read_stream(ole, "key", "i")

        self.assertFalse(data)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock, patch

import numpy as np
from numpy.testing import assert_array_equal

from txrm2tiff.txrm.v3 import Txrm3


@patch.multiple(Txrm3, open=MagicMock)
class TestTxrm3(unittest.TestCase):
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.image_info", new={"MosiacMode": [1]})
    def test_is_mosaic_true(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertTrue(txrm.is_mosaic)

    @patch("txrm2tiff.txrm.v3.AbstractTxrm.image_info", new={"MosiacMode": [0]})
    def test_is_mosaic_false(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertFalse(txrm.is_mosaic)

    def test_is_mosaic_false_if_no_mode(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertFalse(txrm.is_mosaic)

    @patch(
        "txrm2tiff.txrm.v3.AbstractTxrm.image_info",
        new={"ImageWidth": [20], "ImageHeight": [5]},
    )
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.is_mosaic", new=True)
    def test_image_dims(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertEqual(txrm.image_dims, [20, 5])

    @patch(
        "txrm2tiff.txrm.v3.AbstractTxrm.image_info",
        new={"ImageWidth": [18], "ImageHeight": [14]},
    )
    @patch("txrm2tiff.txrm.v3.Txrm3.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.mosaic_dims", new=(6, 7))
    def test_image_dims_mosaic(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertEqual(txrm.image_dims, [3, 2])

    @patch(
        "txrm2tiff.txrm.v3.AbstractTxrm.image_info",
        new={"ImageWidth": [20], "ImageHeight": [5]},
    )
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.is_mosaic", new=True)
    def test_reference_dims(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertEqual(txrm.reference_dims, [20, 5])

    @patch(
        "txrm2tiff.txrm.v3.AbstractTxrm.image_info",
        new={"ImageWidth": [18], "ImageHeight": [14]},
    )
    @patch("txrm2tiff.txrm.v3.Txrm3.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.mosaic_dims", new=(6, 7))
    def test_reference_dims_mosaic(self):
        txrm = Txrm3("test/path/file.txrm")
        self.assertEqual(txrm.reference_dims, [3, 2])

    @patch("txrm2tiff.txrm.v3.AbstractTxrm.read_single_value_from_stream")
    def test_reference_exposure(self, mocked_read_single_value):
        txrm = Txrm3("test/path/file.txrm")
        mocked_read_single_value.return_value = "test"
        self.assertEqual(txrm.reference_exposure, "test")
        mocked_read_single_value.assert_called_once_with("ReferenceData/ExpTime")

    @patch("txrm2tiff.txrm.v3.Txrm3.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v3.Txrm3.reference_dims", new=[2, 3])
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.mosaic_dims", new=[1, 3])
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.extract_reference_data")
    @patch("txrm2tiff.txrm.v3.fallback_image_interpreter")
    def test_extract_reference_image_unknown_dtype(
        self, mocked_fallback_interpreter, mocked_extract_ref_data
    ):
        strict = False
        txrm = Txrm3(
            "test/path/file.txrm",
            load_images=False,
            load_reference=False,
            strict=strict,
        )
        ole = MagicMock()
        txrm.ole = ole
        im_size = np.prod(txrm.reference_dims)
        arr = np.arange(im_size, dtype=np.uint16)
        arr.shape = txrm.reference_dims[::-1]
        arr_flat = arr.flatten()
        arr_bytes = arr_flat.tobytes()
        mocked_extract_ref_data.return_value = arr_bytes
        mocked_fallback_interpreter.return_value = arr_flat

        output = txrm.extract_reference_image()

        mocked_extract_ref_data.assert_called_once_with()
        mocked_fallback_interpreter.assert_called_once_with(arr_bytes, im_size, strict)

        tiled_arr = np.vstack([arr] * 3)
        assert_array_equal(output, tiled_arr)

    @patch("txrm2tiff.txrm.v3.Txrm3.is_mosaic", new=True)
    @patch("txrm2tiff.txrm.v3.Txrm3.reference_dims", new=[2, 3])
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.mosaic_dims", new=[1, 3])
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.extract_reference_data")
    @patch("txrm2tiff.txrm.v3.fallback_image_interpreter")
    def test_extract_reference_image_known_dtype(
        self, mocked_fallback_interpreter, mocked_extract_ref_data
    ):
        strict = False
        txrm = Txrm3(
            "test/path/file.txrm",
            load_images=False,
            load_reference=False,
            strict=strict,
        )
        ole = MagicMock()
        txrm.ole = ole
        im_size = np.prod(txrm.reference_dims)
        arr = np.arange(im_size, dtype=np.uint16)
        arr.shape = txrm.reference_dims[::-1]
        arr_flat = arr.flatten()
        mocked_extract_ref_data.return_value = arr_flat

        output = txrm.extract_reference_image()

        mocked_extract_ref_data.assert_called_once_with()
        mocked_fallback_interpreter.assert_not_called()

        tiled_arr = np.vstack([arr] * 3)
        assert_array_equal(output, tiled_arr)

    @patch("txrm2tiff.txrm.v3.AbstractTxrm.clear_images")
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.clear_reference")
    @patch("txrm2tiff.txrm.v3.AbstractTxrm.get_images")
    def test_get_output(
        self,
        mocked_get_images,
        mocked_clear_ref,
        mocked_clear_imgs,
    ):
        shape = (3, 4)
        arr = np.arange(np.prod(shape))
        arr.shape = shape
        mocked_get_images.return_value = arr

        strict = False
        txrm = Txrm3(
            "test/path/file.txrm",
            load_images=False,
            load_reference=False,
            strict=strict,
        )

        output = txrm.get_output(load=False, shifts=False, flip=True, clear_images=True)

        mocked_get_images.assert_called_once_with(False)
        mocked_clear_ref.assert_called_once_with()
        mocked_clear_imgs.assert_called_once_with()
        assert_array_equal(output, arr)
        self.assertNotEqual(id(output), id(arr))  # Should be a copy



from abc import ABC, abstractmethod
from datetime import datetime
import re
import logging
import typing
import itertools
import numpy as np
from io import IOBase
from os import PathLike
from numpy.typing import DTypeLike
from olefile import OleFileIO, isOleFile
from pathlib import Path

from ..xradia_properties.enums import XrmDataTypes
from ..xradia_properties.stream_dtypes import streams_dict
from .. import txrm_functions
from .txrm_property import txrm_property


datetime_regex = re.compile(
    r"(\d{2})\/(\d{2})\/(\d{2})?(\d{2}) (\d{2}):(\d{2}):(\d{2})(\.(\d{2}))?"
)
# Groups "1, 2, 4, 5, 6, 7, 9" are "mm, dd, yy, hh, mm, ss, ff"
# (months, days, year, hours, minutes, seconds, decimal of seconds)
# Note that v3 files do not include the century digits.


class AbstractTxrm(ABC):
    def __init__(
        self,
        file: typing.Union[str, PathLike, IOBase, bytes],
        load_images: bool = True,
        load_reference: bool = True,
        strict: bool = False,
    ):
        """Abstract class for wrapping TXRM/XRM files

        Args:
            file (str | PathLike | IOBase | bytes): Path to valid txrm file, a file-like object, or the bytes from an opened file.
            load_images (bool, optional): Load images to memory on init. Defaults to True.
            load_reference (bool, optional): Load reference images (if available) to memory on init. Defaults to True.
            strict (bool, optional): If True, all calls will be treated as strict (raising, not logging, errors). Defaults to False.
        """
        self.ole = None
        self.strict = strict
        self._images = None
        self._reference = None
        self.referenced = False
        self.annotated_image = None
        self.path = None
        self.name = None

        self.open(file)

        if load_images:
            self.load_images()
        if load_reference and self.has_reference:
            self.load_reference()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def open(
        self, file: typing.Optional[typing.Union[str, PathLike, IOBase, bytes]] = None
    ) -> None:
        """Opens txrm file using OleFileIO. Runs on init but can be used to reopen if closed (only PathLike inputs can be reopened without specifying 'file')."""
        if self.file_is_open:
            logging.debug("File %s is already open", self.name)
        else:
            f = None
            if file is None and self.path is not None:
                f = self.path
            else:
                if isinstance(file, (IOBase, bytes)):
                    f = file
                    self.path = None
                    if hasattr(file, "name"):
                        self.name = file.name
                    else:
                        self.name = f"{file.__class__}"
                elif isinstance(file, (str, PathLike)):
                    path = Path(file)
                    if path.exists() and isOleFile(path):
                        self.path = Path(file)
                        self.name = self.path.name
                        f = self.path
                    else:
                        raise IOError("Path is to invalid file")
                else:
                    raise TypeError("Invalid type for argument file")
            if f is not None:
                logging.debug("Opening file %s", self.name)
                self.ole = OleFileIO(f)
            else:
                raise IOError("'%s' is not a valid xrm/txrm file" % self.name)
            if self.ole.fp is None:
                raise IOError("'%s' failed to open for unknown reasons" % self.name)

    def close(self) -> None:
        """Closes txrm file. Can be reopened using open_file."""
        if not self.file_is_open:
            logging.info("File %s is already closed", self.name)
            return
        logging.debug("Closing file %s", self.name)
        self.ole.close()

    def clear_all(self):
        for name in self.properties.keys():
            delattr(self, f"_{name}")

    @property
    def file_is_open(self) -> bool:
        return (
            self.ole is not None and self.ole.fp is not None and not self.ole.fp.closed
        )

    def has_stream(self, key: str) -> typing.Optional[bool]:
        if not self.file_is_open:
            logging.error("Cannot check stream from closed file")
            return None
        return self.ole.exists(key)

    def list_streams(self) -> typing.List[typing.Optional[str]]:
        if not self.file_is_open:
            logging.error("Cannot list streams when file is closed")
            return []
        return [
            "/".join(stream)
            for stream in self.ole.listdir(streams=True, storages=False)
        ]

    def read_stream(
        self,
        key: str,
        dtype: typing.Optional[DTypeLike] = None,
        strict: typing.Optional[bool] = None,
    ) -> typing.List[typing.Any]:
        if not self.file_is_open:
            logging.error("Cannot get stream from closed file")
            if self.strict:
                raise IOError("Cannot get stream from closed file")
            return None
        if strict is None:
            strict = self.strict
        if dtype is None:
            dtype = streams_dict.get(key)
            if dtype is None:
                logging.error("Stream does not have known dtype, one must be specified")
                if self.strict:
                    raise IOError(
                        "Stream does not have known dtype, one must be specified"
                    )
        return txrm_functions.read_stream(self.ole, key, dtype, strict)

    def read_single_value_from_stream(
        self, key: str, idx: int = 0, dtype: typing.Optional[DTypeLike] = None
    ) -> typing.Optional[typing.Any]:
        val = self.read_stream(key, dtype)
        if val is None or len(val) <= idx:
            return None
        return val[idx]

    def load_images(self) -> None:
        try:
            if not self.file_is_open:
                logging.error("Cannot read image while file is closed")
            logging.info("Getting images from the file")
            self._images = self.extract_images()
            self.referenced = False
        except IOError:
            logging.error("Failed to load images", exc_info=True)
            if self.strict:
                raise

    def get_images(self, load: bool = True) -> typing.Optional[np.ndarray]:
        """Get images from file (numpy ndarray with shape [idx, y, x]).

        Args:
            load (bool, optional): try to load image if not already loaded. Defaults to True.

        Returns:
            Images as 3D numpy array with shape [idx, y, x], or None if not loaded.
        """
        if self._images is None and load:
            self.load_images()
        return self._images

    def clear_images(self) -> None:
        self._images = None

    def clear_reference(self) -> None:
        self._reference = None

    def load_reference(self) -> None:
        try:
            self._reference = self.extract_reference_image()
        except KeyError:
            if self.strict:
                raise
            logging.warning("No reference is available to load")
        except Exception:
            if self.strict:
                raise
            logging.error("Error occurred extracting reference image", exc_info=True)


    def get_reference(self, load: bool = True) -> typing.Optional[np.ndarray]:
        """Get images from file (numpy ndarray with shape [idx, y, x]).

        Args:
            load (bool, optional): Try to load image if not already loaded. Defaults to True.
            rescaled_by_exp (bool, optional): Rescale reference by relative exposure ()

        Returns:
            Images as 2D numpy array with shape [y, x], or None if not loaded.
        """
        if self._reference is None and load:
            self.load_reference()
        return self._reference

    def _extract_single_image(self, image_num: int, strict: bool = False) -> np.ndarray:
        try:
            # Read the images - They are stored in the txrm as ImageData1 ...
            # Each folder contains 100 images 1-100, 101-200
            img_key = f"ImageData{int(np.ceil(image_num / 100.0))}/Image{image_num}"
            imgdata = None
            if not self.has_stream(img_key):
                raise KeyError("Stream '%s' does not exist" % img_key)
            img_stream_bytes = self.ole.openstream(img_key).getvalue()
            if self.image_dtype is not None:
                try:
                    imgdata = txrm_functions.get_stream_from_bytes(
                        img_stream_bytes, dtype=self.image_dtype.value
                    )
                except Exception:
                    logging.error(
                        "Image could not be extracted using expected dtype '%s'",
                        self.image_dtype,
                    )
            if imgdata is None:  # if dtype was not given or that method failed
                img_size = np.prod(self.shape)
                imgdata = txrm_functions.fallback_image_interpreter(
                    img_stream_bytes, img_size, self.strict
                )
            if imgdata.size > 0:
                imgdata.shape = self.shape  # Resize if not empty
            return imgdata
        except Exception:
            if strict:
                raise
            return np.asarray([])

    def extract_images(
        self,
        start: int = 1,
        end: int = None,
    ) -> np.ndarray:
        """
        Extract only specified range of images using the arguments start and num_images

        Args:
            start (int, optional): First image to extract (from 1). Defaults to 1.
            end (int, optional): Last image to extract (end = images taken if end > images taken). Defaults to extracting all images taken.
            strict (bool, optional): Log exceptions if False, raise errors if True. Defaults to False.

        Raises:
            AttributeError: No images found

        Returns:
            numpy.ndarray: Array of images with shape [idx, y, x]
        """
        try:
            images_taken = self.image_info.get("ImagesTaken", [0])[0]

            if start < 1 or start > images_taken:
                raise ValueError("Cannot extract Image%d as it does not exist" % start)

            remaining_imgs = images_taken - start + 1

            if end is None:
                # Defaults to all extracting all images
                end = images_taken
            elif end > images_taken:
                logging.warning(
                    "Cannot return up to image %d as only %d images exist, returning images %d to %d",
                    end,
                    images_taken,
                    start,
                    remaining_imgs,
                )
                end = remaining_imgs

            if np.prod(self.shape) * images_taken == 0:
                raise AttributeError("No images found")
            # Iterates through images until the number of images taken
            # lambda check has been left in in case stream is wrong
            images = (
                self._extract_single_image(i, strict=False)
                for i in range(start, end + 1)
            )
            return np.asarray(
                tuple(itertools.takewhile(lambda image: image.size > 1, images))
            )
        except Exception:
            if self.strict:
                raise
            logging.error("Error occurred extracting images", exc_info=True)
            return np.asarray([])

    def get_central_image(self) -> np.ndarray:
        """Returns central image of an odd count stack, or the first of the two central images if an even number"""
        images_taken = self.image_info["ImagesTaken"][0]
        central_img = 1 + images_taken // 2  # First image is 1
        return self.get_single_image(central_img)

    def extract_reference_data(
        self,
    ) -> typing.Optional[typing.Union[np.ndarray, bytes]]:
        """Returns 2D numpy array of reference image in row-column order"""
        # Read the reference image.
        # Reference info is stored under 'ReferenceData/...'
        if not self.has_reference:
            raise KeyError("ReferenceData/Image does not exist")
        ref_stream_bytes = self.ole.openstream("ReferenceData/Image").getvalue()
        if not self.reference_dtype:
            logging.error(
                "Image could not be extracted using expected dtype '%s', returning bytes",
                self.reference_dtype,
            )
            return ref_stream_bytes
        return txrm_functions.get_stream_from_bytes(
            ref_stream_bytes, dtype=self.reference_dtype.value
        )

    @txrm_property(fallback=[])
    def shape(self) -> typing.List[int]:
        """Shape of the stored image(s) in the order (y row, x col). This is the shape of the whole mosaic if a stitched image is stored."""
        if self._images is None:
            return [self.image_info["ImageHeight"][0], self.image_info["ImageWidth"][0]]
        shape = self._images.shape
        if len(shape) > 2:
            shape = self._image[0].shape
        return shape

    @txrm_property(fallback=dict())
    def image_info(self):
        return txrm_functions.get_image_info_dict(
            self.ole, ref=False, strict=self.strict
        )

    @txrm_property(fallback=dict())
    def reference_info(self):
        return txrm_functions.get_image_info_dict(
            self.ole, ref=True, strict=self.strict
        )

    @txrm_property(fallback=dict())
    def position_info(self):
        return txrm_functions.get_position_dict(self.ole)

    @txrm_property(fallback=None)
    def version(self):
        return txrm_functions.get_file_version(self.ole, strict=self.strict)

    @txrm_property(fallback=0)
    def zero_angle_index(self) -> float:
        angles = self.image_info.get("Angles", [])
        if len(self.exposures) <= 1 or len(np.unique(angles)) <= 1:
            # If only a single (or no) exposure or sample theta is consitent, return 0
            return 0
        # Return index of angle closest to 0
        return np.array([abs(angle) for angle in angles]).argmin()

    @txrm_property(fallback=[])
    def mosaic_dims(self) -> typing.List[int]:
        """Returns List of mosaic dims [x columns, y rows]"""
        return [
            self.image_info["MosiacColumns"][0],
            self.image_info["MosiacRows"][0],
        ]

    @txrm_property(fallback=[])
    def energies(self) -> typing.List[float]:
        energies = self.image_info["Energy"]
        if not np.sum(energies):
            # position_info includes units
            energies = self.position_info["Energy"][0]
        if np.sum(energies):
            return energies
        raise ValueError("Could not get energies")

    @txrm_property(fallback=[])
    def exposures(self) -> typing.List[float]:
        """Returns list of exposures"""
        if "ExpTimes" in self.image_info:
            return self.image_info["ExpTimes"]
        elif "ExpTime" in self.image_info:
            return self.image_info["ExpTime"]
        elif self.strict:
            raise KeyError("No exposure time available in ole file.")
        logging.error("No exposure time available in ole file.")
        return []

    @txrm_property(fallback=None)
    def has_reference(self) -> bool:
        return self.has_stream("ReferenceData/Image")

    @txrm_property(fallback=[])
    def datetimes(self) -> typing.List[datetime]:
        dates = []
        for date_str in self.image_info["Date"]:
            m = datetime_regex.search(date_str)
            if m:  # Ignore out any random characters that aren't dates
                if m.group(3):  # Century digits
                    pattern = r"%m%d%Y%H%M%S"
                else:
                    pattern = r"%m%d%y%H%M%S"
                if m.group(9):  # Fraction of a second
                    pattern += r"%f"
                date_str = "".join([d for d in m.group(1, 2, 3, 4, 5, 6, 7, 9) if d])
                dates.append(datetime.strptime(date_str, pattern))
        return dates

    def get_single_image(self, idx: int) -> np.ndarray:
        """Get a single image (from memory if images are loaded). idx starts from 1."""
        images = self.get_images(load=False)
        if images is not None:
            return images[idx - 1]
        else:
            return self._extract_single_image(idx, strict=self.strict)

    @txrm_property(fallback=None)
    def image_dtype(self) -> XrmDataTypes:
        return XrmDataTypes.from_number(
            self.read_single_value_from_stream("ImageInfo/DataType")
        )

    @txrm_property(fallback=None)
    def reference_dtype(self) -> XrmDataTypes:
        return XrmDataTypes.from_number(
            self.read_single_value_from_stream("ReferenceData/DataType")
        )

    @txrm_property(fallback=[])
    def output_shape(self) -> typing.List[int]:
        """
        Returns shape that the output file will be in numpy ordering [idx, y, x]

        Functions overriding this function need to use the wrapper txrm_property
        """
        dims = self.image_dims
        if self.is_mosaic:
            dims = [dim * m_dim for dim, m_dim in zip(dims, self.mosaic_dims)]
            stack = 1
        else:
            stack = self.image_info["ImagesTaken"][0]
        return [stack, *dims[::-1]]

    @txrm_property(fallback=None)
    @abstractmethod
    def is_mosaic(self) -> bool:
        """Functions overriding this function need to use the wrapper txrm_property"""
        raise NotImplementedError

    @txrm_property(fallback=[])
    @abstractmethod
    def image_dims(self) -> typing.List[int]:
        """
        Dimensions of a single image

        Functions overriding this function need to use the wrapper txrm_property
        """
        raise NotImplementedError

    @txrm_property(fallback=[])
    @abstractmethod
    def reference_dims(self) -> typing.List[int]:
        """
        Dimensions of the reference image

        Functions overriding this function need to use the wrapper txrm_property
        """
        raise NotImplementedError

    @txrm_property(fallback=None)
    @abstractmethod
    def reference_exposure(self) -> float:
        """Functions overriding this function need to use the wrapper txrm_property"""
        raise NotImplementedError

    @abstractmethod
    def extract_reference_image(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_output(
        self,
        load: bool = False,
        shifts: bool = False,
        flip: bool = False,
        clear_images: bool = True,
    ) -> typing.Optional[np.ndarray]:
        raise NotImplementedError



from datetime import datetime
import unittest
from unittest.mock import patch, MagicMock

import numpy as np

from txrm2tiff.utils.metadata import create_ome_metadata


class TestMetadata(unittest.TestCase):
    def test_metadata_created_correctly(self):
        filename = "test_file.ext"
        dims = (45, 40, 1)
        mosaic_rows = 2
        mosaic_cols = 3

        txrm = MagicMock()

        txrm.output_shape = dims[::-1]
        exposures = [2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
        txrm.exposures = exposures
        txrm.image_info = {
            "PixelSize": [0.005],
            "XPosition": [7.5, 22.5, 37.5] * 2,
            "YPosition": [10, 10, 10, 20, 20, 20],
            "ZPosition": [1, 1, 1, 1, 1, 1],
        }
        txrm.datetimes = [datetime(2021, 12, 30, 23, 55, 59)]
        txrm.mosaic_dims = [mosaic_cols, mosaic_rows]
        txrm.is_mosaic = False

        ome_metadata = create_ome_metadata(txrm, filename)
        self.assertEqual(ome_metadata.image().Pixels.get_SizeX(), dims[0])
        self.assertEqual(ome_metadata.image().Pixels.get_SizeY(), dims[1])
        self.assertEqual(ome_metadata.image().Pixels.get_SizeT(), dims[2])
        self.assertEqual(
            ome_metadata.image().get_AcquisitionDate(), "2021-12-30T23:55:59"
        )

    def test_mosaic_exposure_averaged(self):
        filename = "test_file.ext"
        dims = (45, 40, 1)
        mosaic_rows = 2
        mosaic_cols = 3

        txrm = MagicMock()

        txrm.output_shape = dims[::-1]
        exposures = [
            2.0,
            3.0,
            4.0,
            5.0,
            0.0,
            0.0,
        ]  # 0 exposures should be ignored, if they exist, as these will be interrupted frames
        txrm.exposures = exposures
        txrm.image_info = {
            "PixelSize": [0.005],
            "XPosition": [7.5, 22.5, 37.5] * 2,
            "YPosition": [10, 10, 10, 20, 20, 20],
            "ZPosition": [1, 1, 1, 1, 1, 1],
        }
        txrm.datetimes = [datetime(2021, 12, 30, 23, 55, 59)]
        txrm.mosaic_dims = [mosaic_cols, mosaic_rows]
        txrm.is_mosaic = True

        expected_exposure = 3.5

        ome_metadata = create_ome_metadata(txrm, filename)
        self.assertEqual(
            ome_metadata.image().Pixels.Plane(0).get_ExposureTime(), expected_exposure
        )

    def test_mosaic_centre_found_correctly(self):
        filename = "test_file.ext"
        dims = (45, 40, 1)
        pixel_size = 0.005
        mosaic_cols = 3
        mosaic_rows = 2
        exposure_times = [0.0, 0.0, 2.0, 3.0, 4.0, 5.0]
        offset = [3.0, -2.0]
        # Multiplier of 1.e3 required as units from xrm files are micrometres and the output should be in nanometres
        expected_centre = (
            (22.5 + offset[0]) * pixel_size * 1.0e3,
            (20.0 + offset[1]) * pixel_size * 1.0e3,
        )

        txrm = MagicMock()

        txrm.output_shape = dims[::-1]
        txrm.exposures = exposure_times
        # This should only need the coords of the first frame as mosaic may not complete
        txrm.image_info = {
            "PixelSize": [pixel_size],
            "XPosition": [
                (7.5 + offset[0]) * pixel_size,
                0,
                0,
                0,
                0,
                0,
            ],
            "YPosition": [
                (10 + offset[1]) * pixel_size,
                0,
                0,
                0,
                0,
                0,
            ],
            "ZPosition": [1, 1, 1, 1, 1, 1],
        }
        txrm.datetimes = [datetime(2021, 12, 30, 23, 55, 59)]
        txrm.mosaic_dims = [mosaic_cols, mosaic_rows]
        txrm.is_mosaic = True

        ome_metadata = create_ome_metadata(txrm, filename)
        plane = ome_metadata.image().Pixels.Plane(0)
        ome_centre = (float(plane.get_PositionX()), float(plane.get_PositionY()))
        [
            self.assertAlmostEqual(ome, expected)
            for ome, expected in zip(ome_centre, expected_centre)
        ]



import logging
import re
import typing
import numbers
import numpy as np
import olefile as of
import numpy.typing as npt

from .. import xradia_properties as xp

hex_pattern = re.compile(r"[^\x20-\x7e]+")


def read_stream(
    ole: of.OleFileIO,
    key: str,
    dtype: typing.Optional[typing.Union[xp.XrmDataTypes, npt.DTypeLike, None]] = None,
    strict: bool = False,
) -> typing.List[typing.Any]:
    """Reads and returns list containing stream specified by key, decoded as dtype"""
    try:
        if dtype is None:
            dtype = xp.streams_dict.get(key)
            if dtype is None:
                raise TypeError("No known data type found, one must be specified.")
        if isinstance(dtype, xp.XrmDataTypes):
            # streams_dict returns XrmDataTypes
            dtype = dtype.value
        dtype = np.dtype(dtype)  # cast to numpy dtype
        if ole.exists(key):
            if dtype == np.str_:
                return _read_text_stream_to_list(ole, key)
            return _read_number_stream_to_list(ole, key, dtype)
        raise KeyError("Stream %s does not exist in ole file" % key)
    except Exception:
        if strict:
            raise
        logging.error(
            "Error occurred reading stream '%s' as %s", key, dtype, exc_info=True
        )
        return []


def get_stream_from_bytes(stream_bytes: bytes, dtype: npt.DTypeLike) -> np.ndarray:
    """Converts olefile bytes to np.ndarray of values of type dtype."""
    return np.frombuffer(stream_bytes, dtype)


def _read_number_stream_to_list(
    ole: of.OleFileIO, key: str, dtype: typing.Union[npt.DTypeLike, None]
) -> typing.List[typing.Union[numbers.Number, bytes]]:
    """Reads olefile stream and returns to list of values of type dtype."""
    stream_bytes = ole.openstream(key).getvalue()
    return get_stream_from_bytes(stream_bytes, dtype).tolist()


def _read_text_stream_to_list(ole: of.OleFileIO, key: str) -> typing.List[str]:
    """
    Returns list of strings.

    The byte string is decoded then split by hex bytes.
    This is because text is stored in byte strings with seemingly random hex.

    Splitting by hex keeps real spaces while splitting entries (such as with Date).
    """
    if ole.exists(key):
        byte_str = ole.openstream(key).read()
        return [
            s for s in hex_pattern.split(byte_str.decode("ascii", errors="ignore")) if s
        ]


def get_image_info_dict(
    ole: of.OleFileIO, ref: bool = False, strict: bool = False
) -> typing.Dict[str, typing.List[typing.Any]]:
    """
    Reads a selection of useful ImageInfo streams from an XRM/TXRM file.

    Args:
        ole (OleFileIO): Opened XRM/TXRM file.
        ref (bool, optional): If True, returns "ReferenceData/ImageInfo/*". Defaults to False.

    Returns:
        Dict[str, List[Any]]: Dictionary containing the stream keys and their value(s), which are stored in a list.
    """
    image_info = {}
    if ref:
        info_dict = xp.stream_dtypes.ref_image_info_dict
    else:
        info_dict = xp.stream_dtypes.image_info_dict
    for key, dtype in info_dict.items():
        try:
            dict_key = key.split("/")[-1]  # No need for the whole stream
            image_info[dict_key] = read_stream(ole, key, dtype, True)
        except ValueError:
            if strict:
                raise
            logging.error("Invalid data type %s for %s", dtype, key)
        except KeyError:
            logging.debug("Key %s does not exist", key)
    return image_info


def get_position_dict(
    ole: of.OleFileIO,
) -> typing.Dict[str, typing.Tuple[typing.Union[typing.List, str]]]:
    """
    Gets dictionary of motor positions.

    Args:
        ole (of.OleFileIO)

    Returns:
        typing.Dict[str, typing.Tuple[typing.Union[typing.List, str]]]: all positions for each motor in the form: {motor name string: ([values], unit string)}.
    """
    tmp = {}
    for key, dtype in xp.stream_dtypes.position_info_dict.items():
        tmp[key] = read_stream(ole, key, dtype)

    num_axes = tmp["PositionInfo/TotalAxis"][0]
    positions_dict = {}
    # Motor positions are stored in list of all values for all images. It lists
    # all axes for each frame (in that order).
    for i in range(num_axes):
        positions_dict[tmp["PositionInfo/AxisNames"][i]] = (
            tmp["PositionInfo/MotorPositions"][i::num_axes],
            tmp["PositionInfo/AxisUnits"][i],
        )
    return positions_dict


def get_file_version(ole: of.OleFileIO, strict: bool = False) -> typing.Optional[float]:
    v = read_stream(ole, "Version", xp.XrmDataTypes.XRM_FLOAT, strict=strict)
    if v:
        return v[0]
    return None



import unittest
import numpy as np

from txrm2tiff.utils.functions import conditional_replace, convert_to_int


class TestUtilsFunctions(unittest.TestCase):
    def test_conditional_replace(self):
        array = np.repeat(np.arange(0, 9000, 0.1).reshape(300, 300), 10, 0)
        threshold = 100

        conditional_replace(array, np.nan, lambda x: x < threshold)

        self.assertEqual(
            np.nanmin(array),
            threshold,
            msg="Array values below threshold {} have not been replaced with nan".format(
                threshold
            ),
        )

    def test_convert_to_int(self):
        for i in range(0, 320000):
            self.assertEqual(convert_to_int(float(i)), i)

    def test_convert_to_int_fails(self):
        with self.assertRaises(ValueError):
            convert_to_int(3.2)



import unittest
from parameterized import parameterized

import logging
import os
from pathlib import Path
from time import time, sleep
from shutil import rmtree
from io import IOBase
import numpy as np
import tifffile as tf

import txrm2tiff


visit_path = Path(
    "/dls/science/groups/das/ExampleData/B24_test_data/data/2019/cm98765-1"
)
raw_path = visit_path / "raw"
xm10_path = raw_path / "XMv10"
xm13_path = raw_path / "XMv13"

test_files = [
    (xm13_path / "Xray_mosaic_v13.xrm",),
    (xm13_path / "Xray_mosaic_v13_interrupt.xrm",),
    (xm13_path / "Xray_mosaic_7x7_v13.xrm",),
    (xm13_path / "Xray_single_v13.xrm",),
    (xm13_path / "tomo_v13_full.txrm",),
    (xm13_path / "tomo_v13_full_noref.txrm",),
    (xm13_path / "tomo_v13_interrupt.txrm",),
    (xm13_path / "VLM_mosaic_v13.xrm",),
    (xm13_path / "VLM_mosaic_v13_interrupt.xrm",),
    (xm13_path / "VLM_grid_mosaic_large_v13.xrm",),
    (xm10_path / "12_Tomo_F4D_Area1_noref.txrm",),
    (xm10_path / "VLM_mosaic.xrm",),
    (xm10_path / "test_tomo2_e3C_full.txrm",),
    (xm10_path / "Xray_mosaic_F5A.xrm",),
]

@unittest.skipUnless(visit_path.exists(), "dls paths cannot be accessed")
class TestTxrm2TiffWithFiles(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.visit_path = visit_path
        cls.raw_path = raw_path
        cls.processed_path = cls.visit_path / "processed"
        if cls.processed_path.exists():
            cleanup_timeout = 40
            # rmtree is asynchronous, so a wait may be required:
            rmtree(cls.processed_path, ignore_errors=True)
            start = time()
            while (
                os.path.exists(cls.processed_path)
                and (time() - start) < cleanup_timeout
            ):
                sleep(1)

    def setUp(self):
        self.processed_path.mkdir(exist_ok=True)
        self.assertTrue(
            self.processed_path.exists(), msg="Processed folder not correctly created"
        )

    def tearDown(self):
        cleanup_timeout = 40
        # rmtree is asynchronous, so a wait may be required:
        rmtree(self.processed_path, ignore_errors=True)
        start = time()
        while (
            os.path.exists(self.processed_path) and (time() - start) < cleanup_timeout
        ):
            sleep(1)

        self.assertFalse(
            self.processed_path.exists(), msg="Processed folder not correctly removed"
        )

    @parameterized.expand(test_files)
    def test_convert_and_save(self, test_file):
        logging.debug("Running with file %s", test_file)
        output_path = self.processed_path / test_file.relative_to(
            self.raw_path
        ).with_suffix(".ome.tiff")

        # Make processed/ subfolders:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        txrm2tiff.convert_and_save(test_file, output_path=output_path)

        self.assertTrue(output_path.is_file())

    def test_convert_and_save_with_dtype(self):
        test_file = test_files[0][0]
        dtypes = ["uint16", "float32", "float64", np.float32, np.float64, np.uint16]
        logging.debug("Running with file %s", test_file)

        for dtype in dtypes:
            output_path = self.processed_path / (
                test_file.parent / f"{test_file.stem}_{dtype}.ome.tiff"
            ).relative_to(self.raw_path)
            # Make processed/ subfolders:
            output_path.parent.mkdir(parents=True, exist_ok=True)

            txrm2tiff.convert_and_save(
                test_file, output_path=output_path, data_type=dtype
            )

            self.assertTrue(output_path.exists())
            with tf.TiffFile(str(output_path)) as tif:
                a = tif.asarray()
            self.assertEqual(
                a.dtype, np.dtype(dtype), msg=f"dtype is {a.dtype} not {dtype}"
            )

    @parameterized.expand(test_files)
    def test_convert_with_txrm_class(self, test_file):
        logging.debug("Running with file %s", test_file)
        output_path = self.processed_path / test_file.relative_to(
            self.raw_path
        ).with_suffix(".ome.tiff")

        # Make processed/ subfolders:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with txrm2tiff.open_txrm(test_file, strict=True) as txrm:
            txrm.apply_reference()
            self.assertTrue(txrm.save_images(output_path))

        self.assertTrue(output_path.is_file())

    def test_convert_with_txrm_class_with_dtype(self):
        test_file = test_files[0][0]
        dtypes = ["uint16", "float32", "float64", np.float32, np.float64, np.uint16]
        logging.debug("Running with file %s", test_file)

        with txrm2tiff.open_txrm(test_file, strict=True) as txrm:
            if txrm.has_reference:
                txrm.apply_reference()
            for dtype in dtypes:
                output_path = self.processed_path / (
                    test_file.parent / f"{test_file.stem}_{dtype}.ome.tiff"
                ).relative_to(self.raw_path)
                # Make processed/ subfolders:
                output_path.parent.mkdir(parents=True, exist_ok=True)

                self.assertTrue(txrm.save_images(output_path, dtype))

                self.assertTrue(output_path.exists())
                with tf.TiffFile(str(output_path)) as tif:
                    a = tif.asarray()
                self.assertEqual(
                    a.dtype, np.dtype(dtype), msg=f"dtype is {a.dtype} not {dtype}"
                )

    def test_convert_with_txrm_class_using_open_file_buffer(self):
        test_file = test_files[0][0]
        logging.debug("Running with file %s", test_file)

        with test_file.open("rb") as f:
            self.assertTrue(isinstance(f, IOBase))

            with txrm2tiff.open_txrm(f, strict=True) as txrm:
                if txrm.has_reference:
                    txrm.apply_reference()
                output_path = self.processed_path / (
                    test_file.parent / f"{test_file.stem}.ome.tiff"
                ).relative_to(self.raw_path)
                # Make processed/ subfolders:
                output_path.parent.mkdir(parents=True, exist_ok=True)

                self.assertTrue(txrm.save_images(output_path))

        self.assertTrue(output_path.is_file())

    def test_convert_with_txrm_class_using_bytes(self):
        test_file = test_files[0][0]
        logging.debug("Running with file %s", test_file)

        with test_file.open("rb") as f:
            self.assertTrue(isinstance(f, IOBase))
            bytestring = f.read()

        with txrm2tiff.open_txrm(bytestring, strict=True) as txrm:
            if txrm.has_reference:
                txrm.apply_reference()
            output_path = self.processed_path / (
                test_file.parent / f"{test_file.stem}.ome.tiff"
            ).relative_to(self.raw_path)
            # Make processed/ subfolders:
            output_path.parent.mkdir(parents=True, exist_ok=True)

            self.assertTrue(txrm.save_images(output_path))

        self.assertTrue(output_path.is_file())



import unittest
from unittest.mock import MagicMock, patch
from numpy.testing import assert_array_equal

import numpy as np

from txrm2tiff.txrm.shifts_mixin import ShiftsMixin


class TestShifts(unittest.TestCase):
    def test_apply_shifts_to_images_x_shift(self):
        base = range(5)
        arr = np.vstack([base, base, base, base])[np.newaxis, :, :]

        shifter = ShiftsMixin()
        shifter.read_stream = MagicMock(side_effect=[[2], [0]])
        output = shifter.apply_shifts_to_images(images=arr)

        expected_base = [3, 4, 0, 1, 2]
        expected_output = np.vstack(
            [expected_base, expected_base, expected_base, expected_base]
        )[np.newaxis, :, :]
        assert_array_equal(output, expected_output)

    def test_apply_shifts_to_images_y_shift(self):
        base = range(5)
        arr = np.vstack([base, base, base, base]).transpose()[np.newaxis, :, :]

        shifter = ShiftsMixin()
        shifter.read_stream = MagicMock(side_effect=[[0], [2]])
        output = shifter.apply_shifts_to_images(images=arr)

        expected_base = [3, 4, 0, 1, 2]
        expected_output = np.vstack(
            [expected_base, expected_base, expected_base, expected_base]
        ).transpose()[np.newaxis, :, :]
        assert_array_equal(output, expected_output)

    def test_apply_shifts_to_images_xy_shift(self):
        base = np.arange(0, 5)
        width = len(base)
        arr = np.vstack(
            [
                base,
                base + width,
                base + 2 * width,
                base + 3 * width,
                base + 4 * width,
                base + 5 * width,
            ]
        )[np.newaxis, :, :]

        shifter = ShiftsMixin()
        shifter.read_stream = MagicMock(side_effect=[[3], [2]])
        output = shifter.apply_shifts_to_images(images=arr)

        expected_output = np.asarray(
            [
                [
                    [22, 23, 24, 20, 21],
                    [27, 28, 29, 25, 26],
                    [2, 3, 4, 0, 1],
                    [7, 8, 9, 5, 6],
                    [12, 13, 14, 10, 11],
                    [17, 18, 19, 15, 16],
                ]
            ]
        )
        assert_array_equal(output, expected_output)



from numbers import Number
from typing import Callable
from numpy import ndarray


def convert_to_int(value: Number) -> int:
    if int(value) == float(value):
        return int(value)
    raise ValueError(f"Value '{value}' cannot be converted to an integer")


def conditional_replace(
    array: ndarray, replacement: Number, condition_func: Callable
) -> ndarray:
    array[condition_func(array)] = replacement



import unittest
from unittest.mock import patch, MagicMock, call
from parameterized import parameterized

from pathlib import Path
from tempfile import TemporaryDirectory
from random import choice, randint
from string import ascii_letters

from txrm2tiff.main import (
    convert_and_save,
    _batch_convert_files,
    _convert_and_save,
    _set_output_suffix,
    _convert_file,
    _decide_output_path,
    _find_files,
)


from txrm2tiff.txrm.v3 import Txrm3
from txrm2tiff.txrm.v5 import Txrm5


def generate_random_strings(return_num: int):
    for i in range(return_num):
        yield f"{''.join([choice(ascii_letters) for _ in range(10)])}"


class TestRun(unittest.TestCase):
    def test_set_output_suffix(self):
        txrm_output = _set_output_suffix(Path("file.txrm"))
        self.assertEqual("file.ome.tiff", str(txrm_output))
        xrm_output = _set_output_suffix(Path("file.xrm"))
        self.assertEqual("file.ome.tiff", str(xrm_output))
        txrm_output2 = _set_output_suffix(Path("file.extension"), ".txrm")
        self.assertEqual("file.ome.tiff", str(txrm_output2))
        xrm_output2 = _set_output_suffix(Path("file.extension"), ".xrm")
        self.assertEqual("file.ome.tiff", str(xrm_output2))
        with self.assertRaises(NameError):
            _set_output_suffix(Path("file.bad_extension"), None)

    def test_set_output_suffix(self):
        txrm_output = _set_output_suffix(Path("file.txrm"))
        xrm_output = _set_output_suffix(Path("file.xrm"))
        self.assertEqual("file.ome.tiff", str(txrm_output))
        self.assertEqual("file.ome.tiff", str(xrm_output))

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    @patch("txrm2tiff.main.open_txrm")
    def test_convert_and_save(self, TxrmClass, mocked_open_txrm):
        input_filepath = Path("test_file.txrm")

        txrm = MagicMock(auto_spec=TxrmClass)
        txrm.path = input_filepath
        txrm.name = input_filepath.name
        mocked_open_txrm.return_value.__enter__.return_value = txrm
        _convert_and_save(input_filepath, None, None, False, False, None, False)
        mocked_open_txrm.assert_called_with(input_filepath)
        txrm.save_images.assert_called_with(
            input_filepath.with_suffix(".ome.tiff"),
            None,
            flip=False,
            shifts=False,
            mkdir=True,
        )

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    def test_convert_file(self, TxrmClass):
        input_filepath = Path("test_file.txrm")
        txrm = MagicMock(auto_spec=TxrmClass)
        txrm.path = input_filepath
        txrm.name = input_filepath.name
        _convert_file(txrm, None, False, False)
        txrm.apply_reference.assert_called_once_with(
            None
        )  # Called with None applies internal reference
        txrm.annotate.assert_not_called()

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    def test_convert_file_ignore_reference(self, TxrmClass):
        input_filepath = Path("test_file.txrm")
        txrm = MagicMock(auto_spec=TxrmClass)
        txrm.path = input_filepath
        txrm.name = input_filepath.name
        _convert_file(txrm, None, True, False)
        txrm.apply_reference.assert_not_called()
        txrm.annotate.assert_not_called()

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    def test_convert_file_custom_reference_overrules_ignore_reference(self, TxrmClass):
        input_filepath = Path("test_file.txrm")
        custom_ref = Path("path/custom_ref")
        txrm = MagicMock(auto_spec=TxrmClass)
        txrm.name = input_filepath.name
        _convert_file(txrm, custom_ref, True, False)
        txrm.apply_reference.assert_called_once_with(custom_ref)
        txrm.annotate.assert_not_called()

    @parameterized.expand([(Txrm3,), (Txrm5,)])
    def test_convert_file_with_annotation(self, TxrmClass):
        input_filepath = Path("test_file.txrm")
        txrm = MagicMock(auto_spec=TxrmClass)
        txrm.path = input_filepath
        txrm.name = input_filepath.name
        _convert_file(txrm, None, True, True)
        txrm.annotate.assert_called_once_with()

    @parameterized.expand([(".txrm", ".ome.tiff"), (".xrm", ".ome.tiff")])
    def test_decide_output_path(self, input_suffix, output_suffix):
        filepath = Path("./path/to/thisisafile.ext")
        output = _decide_output_path(filepath.with_suffix(input_suffix), None)
        self.assertEqual(output, filepath.with_suffix(output_suffix))

    @parameterized.expand([(".txrm", ".ome.tiff"), (".xrm", ".ome.tiff")])
    def test_decide_output_path_with_output_dir(self, input_suffix, output_suffix):
        filepath = Path("./path/to/thisisafile.ext")
        output_dir = Path("./a/different/path")
        output = _decide_output_path(filepath.with_suffix(input_suffix), output_dir)
        self.assertEqual(
            output, (output_dir / filepath.name).with_suffix(output_suffix)
        )

    @parameterized.expand([(".txrm"), (".xrm")])
    def test_decide_output_path_with_output_path(self, input_suffix):
        filepath = Path("./path/to/thisisafile.ext")
        output_path = Path("./a/different/path.ext")
        output = _decide_output_path(filepath.with_suffix(input_suffix), output_path)
        self.assertEqual(output, output_path)

    def test_find_files(self):
        with TemporaryDirectory(prefix="test_finds_files_", dir=".") as tmpdir:
            test_dir = Path(tmpdir)
            txrm_subdir = test_dir / "txrm_subdir"
            txrm_subdir.mkdir()
            xrm_subdir = test_dir / "xrm_subdir"
            xrm_subdir.mkdir()
            bad_subdir = test_dir / "bad_subdir"
            bad_subdir.mkdir()
            num_txrm = 10
            num_xrm = 5
            num_bad = 7
            txrm_files = {
                txrm_subdir / f"{fname}.txrm"
                for fname in generate_random_strings(num_txrm)
            }

            xrm_files = {
                xrm_subdir / f"{fname}.xxrm"
                for fname in generate_random_strings(num_xrm)
            }
            bad_files = {
                bad_subdir / f"{fname}.ext"
                for fname in generate_random_strings(num_bad)
            }
            for f in set().union(txrm_files, xrm_files, bad_files):
                f.touch()

            self.assertEqual(txrm_files.union(xrm_files), set(_find_files(test_dir)))

    @patch("txrm2tiff.main._set_output_suffix")
    @patch("txrm2tiff.main._find_files")
    @patch("txrm2tiff.main._convert_and_save")
    def test_batch_convert_files(
        self, mocked_convert_save, mocked_find_files, mocked_suffix_definer
    ):
        input_dir = Path("./path/to/thisisadir")
        num_txrm = 2
        num_xrm = 3
        txrm_files = [
            input_dir / "txrm_dir" / f"{fname}.txrm"
            for fname in generate_random_strings(num_txrm)
        ]

        xrm_files = [
            input_dir / f"{fname}.xrm" for fname in generate_random_strings(num_xrm)
        ]
        input_files = txrm_files + xrm_files
        output_files = [f.with_suffix(".ome.tiff") for f in txrm_files] + [
            f.with_suffix(".ome.tiff") for f in xrm_files
        ]
        mocked_find_files.return_value = input_files
        mocked_suffix_definer.side_effect = output_files
        _batch_convert_files(input_dir)
        mocked_convert_save.assert_has_calls(
            [
                call(f_in, f_out, None, True, False, None, False, False)
                for f_in, f_out in zip(input_files, output_files)
            ]
        )

    @patch("pathlib.Path.mkdir", MagicMock())
    @patch("txrm2tiff.main._find_files")
    @patch("txrm2tiff.main._convert_and_save")
    def test_batch_convert_files_to_output(
        self,
        mocked_convert_save,
        mocked_find_files,
    ):
        input_dir = Path("./path/to/thisisadir")
        output_dir = Path("./path/to/output_dir")
        num_txrm = 3
        num_xrm = 2
        txrm_files = [
            Path(f"txrm_dir/{fname}.txrm")
            for fname in generate_random_strings(num_txrm)
        ]

        xrm_files = [Path(f"{fname}.xrm") for fname in generate_random_strings(num_xrm)]
        input_files = txrm_files + xrm_files
        output_files = [f.with_suffix(".ome.tiff") for f in txrm_files] + [
            f.with_suffix(".ome.tiff") for f in xrm_files
        ]
        input_paths = [input_dir / s for s in input_files]
        output_paths = [output_dir / s for s in output_files]
        mocked_find_files.return_value = input_paths
        _batch_convert_files(input_dir, output_dir)
        mocked_convert_save.assert_has_calls(
            [
                call(f_in, f_out, None, True, False, None, False, False)
                for f_in, f_out in zip(input_paths, output_paths)
            ]
        )

    @patch("pathlib.Path.exists", MagicMock(return_value=False))
    def test_convert_and_save_non_existant_file(self):
        input_path = "input/path/file.txrm"
        custom_reference = "input/path/custom.xrm"
        output_path = "output/path"
        dtype = "uint8"
        logging_level = "debug"
        with self.assertRaises(IOError):
            convert_and_save(
                input_path,
                output_path,
                custom_reference,
                True,
                False,
                dtype,
                False,
                logging_level,
            )

    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    @patch("txrm2tiff.main.create_logger")
    @patch("txrm2tiff.main._convert_and_save")
    def test_convert_and_save_str_input_invalid_custom_reference(
        self, mocked_convert_and_save, mocked_create_logger
    ):
        input_path = "input/path/file.txrm"
        custom_reference = "input/path/custom.xrm"
        output_path = "output/path"
        dtype = "uint8"
        logging_level = "debug"

        convert_and_save(
            input_path,
            output_path,
            custom_reference,
            True,
            False,
            dtype,
            False,
            False,
            logging_level,
        )

        mocked_create_logger.assert_called_once_with(logging_level)
        mocked_convert_and_save.assert_called_once_with(
            Path(input_path), Path(output_path), None, True, False, dtype, False, False
        )  # Casts paths to Path

    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    @patch("txrm2tiff.main.create_logger")
    @patch("txrm2tiff.main._convert_and_save")
    def test_convert_and_save_Path_input_invalid_custom_reference(
        self, mocked_convert_and_save, mocked_create_logger
    ):
        input_path = Path("input/path/file.txrm")
        custom_reference = Path("input/path/custom.xrm")
        output_path = Path("output/path")
        dtype = "uint8"
        logging_level = "debug"

        convert_and_save(
            input_path,
            output_path,
            custom_reference,
            True,
            False,
            dtype,
            True,
            False,
            logging_level,
        )

        mocked_create_logger.assert_called_once_with(logging_level)
        mocked_convert_and_save.assert_called_once_with(
            input_path, output_path, None, True, False, dtype, True, False
        )

    @patch("pathlib.Path.is_file", MagicMock(return_value=True))
    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    @patch("txrm2tiff.main.create_logger")
    @patch("txrm2tiff.main._convert_and_save")
    def test_convert_and_save_Path_input_custom_reference(
        self, mocked_convert_and_save, mocked_create_logger
    ):
        input_path = Path("input/path/file.txrm")
        custom_reference = Path("input/path/custom.xrm")
        output_path = Path("output/path")
        dtype = "uint8"
        logging_level = "debug"

        convert_and_save(
            input_path,
            output_path,
            custom_reference,
            True,
            False,
            dtype,
            True,
            False,
            logging_level,
        )

        mocked_create_logger.assert_called_once_with(logging_level)
        mocked_convert_and_save.assert_called_once_with(
            input_path, output_path, custom_reference, True, False, dtype, True, False
        )

    @patch("pathlib.Path.is_dir", MagicMock(return_value=True))
    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    @patch("txrm2tiff.main.create_logger")
    @patch("txrm2tiff.main._batch_convert_files")
    def test_convert_and_save_str_input_batch(
        self, mocked_batch_convert_files, mocked_create_logger
    ):
        input_path = "input/path"
        output_path = "output/path"
        dtype = "uint8"
        logging_level = "debug"

        convert_and_save(
            input_path, output_path, None, True, False, dtype, True, True, logging_level
        )

        mocked_create_logger.assert_called_once_with(logging_level)
        mocked_batch_convert_files.assert_called_once_with(
            Path(input_path), Path(output_path), True, False, dtype, True, True
        )

    @patch("pathlib.Path.is_dir", MagicMock(return_value=True))
    @patch("pathlib.Path.exists", MagicMock(return_value=True))
    @patch("txrm2tiff.main.create_logger")
    @patch("txrm2tiff.main._batch_convert_files")
    def test_convert_and_save_Path_input_batch(
        self, mocked_batch_convert_files, mocked_create_logger
    ):
        input_path = Path("input/path")
        output_path = Path("output/path")
        dtype = "uint8"
        logging_level = "debug"

        convert_and_save(
            input_path, output_path, None, True, False, dtype, True, False, logging_level
        )

        mocked_create_logger.assert_called_once_with(logging_level)
        mocked_batch_convert_files.assert_called_once_with(
            input_path, output_path, True, False, dtype, True, False
        )

    @patch("txrm2tiff.main._convert_and_save")
    def test_convert_and_save_with_file(self, mocked_convert):
        with TemporaryDirectory(dir=".") as tmp_in:
            tmp_in_filepath = Path(tmp_in) / f"{randint(0,9999)}.xrm"
            tmp_in_filepath.touch()
            convert_and_save(tmp_in_filepath)
        mocked_convert.assert_called_with(
            tmp_in_filepath, None, None, False, False, None, False, False
        )

    @patch("txrm2tiff.main._batch_convert_files")
    def test_convert_and_save_with_dir(self, mocked_batch_convert):
        with TemporaryDirectory(dir=".") as tmp_in:
            tmp_in_path = Path(tmp_in)
            convert_and_save(tmp_in_path)
        mocked_batch_convert.assert_called_with(
            tmp_in_path, None, False, False, None, False, False
        )
